# -*- coding: utf-8 -*-
# glossary.py
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import logging

import sys

import os
from os.path import (
	split,
	join,
	splitext,
	isfile,
	isdir,
	dirname,
	basename,
	abspath,
)

from time import time as now
import subprocess
import re

import pkgutil
from collections import Counter
from collections import OrderedDict as odict

import io
import gc

from typing import (
	Dict,
	Tuple,
	List,
	Any,
	Optional,
	ClassVar,
	Iterator,
	Callable,
)

from .flags import *
from . import core
from .core import VERSION, userPluginsDir
from .entry_base import BaseEntry
from .entry import Entry, DataEntry
from .sort_stream import hsortStreamList

from .text_utils import (
	fixUtf8,
)
from .os_utils import indir

from .glossary_type import GlossaryType

homePage = "https://github.com/ilius/pyglossary"
log = logging.getLogger("root")

file = io.BufferedReader

try:
	ModuleNotFoundError
except NameError:
	ModuleNotFoundError = ImportError


def get_ext(path: str) -> str:
	return splitext(path)[1].lower()


class Glossary(GlossaryType):
	"""
	Direct access to glos.data is droped
	Use `glos.addEntry(word, defi, [defiFormat])`
		where both word and defi can be list (including alternates) or string
	See help(glos.addEntry)

	Use `for entry in glos:` to iterate over entries (glossary data)
	See help(pyglossary.entry.Entry) for details

	"""

	# Should be changed according to plugins? FIXME
	infoKeysAliasDict = {
		"title": "name",
		"bookname": "name",
		"dbname": "name",
		##
		"sourcelang": "sourceLang",
		"inputlang": "sourceLang",
		"origlang": "sourceLang",
		##
		"targetlang": "targetLang",
		"outputlang": "targetLang",
		"destlang": "targetLang",
		##
		"license": "copyright",
	}
	plugins = {}  # format => pluginModule
	readFunctions = {}
	readerClasses = {}
	writeFunctions = {}
	writerClasses = {}
	formatsDesc = {}
	formatsExt = {}
	formatsReadOptions = {}
	formatsWriteOptions = {}
	formatsOptionsProp = {}
	formatsDepends = {}
	descFormat = {}
	descExt = {}
	extFormat = {}

	readFormats = []
	writeFormats = []
	readExt = []
	writeExt = []
	readDesc = []
	writeDesc = []

	@classmethod
	def loadPlugins(cls: ClassVar, directory: str) -> None:
		"""
		executed on startup.  as name implies, loads plugins from directory
		"""
		log.debug(f"Loading plugins from directory: {directory!r}")
		if not isdir(directory):
			log.error(f"Invalid plugin directory: {directory!r}")
			return

		sys.path.append(directory)
		for _, pluginName, _ in pkgutil.iter_modules([directory]):
			cls.loadPlugin(pluginName)
		sys.path.pop()

	@classmethod
	def getRWOptionsFromFunc(cls, func, format):
		import inspect
		optionsProp = cls.formatsOptionsProp[format]
		sig = inspect.signature(func)
		optNames = []
		for name, param in sig.parameters.items():
			if param.default is inspect._empty:
				if name not in ("self", "glos", "filename", "dirname", "kwargs"):
					log.warning(f"empty default value for {name}: {param.default}")
				continue # non-keyword argument
			if name not in optionsProp:
				log.warning(f"skipping option {name} in plugin {format}")
				continue
			prop = optionsProp[name]
			if prop.disabled:
				log.debug(f"skipping disabled option {name} in {format} plugin")
				continue
			if not prop.validate(param.default):
				log.warning(f"invalid default value for option: {name} = {param.default!r}")
			optNames.append(name)
		return optNames

	@classmethod
	def loadPlugin(cls: ClassVar, pluginName: str) -> None:
		try:
			plugin = __import__(pluginName)
		except ModuleNotFoundError as e:
			log.warning(f"Module {e.name!r} not found, skipping plugin {pluginName!r}")
			return
		except Exception as e:
			log.exception(f"Error while importing plugin {pluginName}")
			return

		if (not hasattr(plugin, "enable")) or (not plugin.enable):
			log.debug(f"Plugin disabled or not a plugin: {pluginName}")
			return

		format = plugin.format

		extensions = plugin.extensions
		# FIXME: deprecate non-tuple values in plugin.extensions
		if isinstance(extensions, str):
			extensions = (extensions,)
		elif not isinstance(extensions, tuple):
			extensions = tuple(extensions)

		if hasattr(plugin, "description"):
			desc = plugin.description
		else:
			desc = f"{format} ({extensions[0]})"

		cls.plugins[format] = plugin
		cls.descFormat[desc] = format
		cls.descExt[desc] = extensions[0]
		for ext in extensions:
			cls.extFormat[ext] = format
		cls.formatsExt[format] = extensions
		cls.formatsDesc[format] = desc
		cls.formatsOptionsProp[format] = getattr(
			plugin,
			"optionsProp",
			{},
		)
		cls.formatsDepends[format] = getattr(
			plugin,
			"depends",
			{},
		)

		hasReadSupport = False
		try:
			Reader = plugin.Reader
		except AttributeError:
			pass
		else:
			for attr in (
				"__init__",
				"open",
				"close",
				"__len__",
				"__iter__",
			):
				if not hasattr(Reader, attr):
					log.error(
						f"Invalid Reader class in {format!r} plugin"
						f", no {attr!r} method"
					)
					break
			else:
				cls.readerClasses[format] = Reader
				hasReadSupport = True
				cls.formatsReadOptions[format] = cls.getRWOptionsFromFunc(
					Reader.open,
					format,
				)

		# ignore "read" function if "Reader" class is present
		if not hasReadSupport:
			try:
				cls.readFunctions[format] = plugin.read
			except AttributeError:
				pass
			else:
				hasReadSupport = True
				cls.formatsReadOptions[format] = cls.getRWOptionsFromFunc(
					plugin.read,
					format,
				)

		if hasReadSupport:
			cls.readFormats.append(format)
			cls.readExt.append(extensions)
			cls.readDesc.append(desc)

		hasWriteSupport = False
		if hasattr(plugin, "Writer"):
			cls.writerClasses[format] = plugin.Writer
			cls.formatsWriteOptions[format] = cls.getRWOptionsFromFunc(
				plugin.Writer.write,
				format,
			)
			hasWriteSupport = True

		if not hasWriteSupport and hasattr(plugin, "write"):
			cls.writeFunctions[format] = plugin.write
			cls.formatsWriteOptions[format] = cls.getRWOptionsFromFunc(
				plugin.write,
				format,
			)
			hasWriteSupport = True

		if hasWriteSupport:
			cls.writeFormats.append(format)
			cls.writeExt.append(extensions)
			cls.writeDesc.append(desc)

		return plugin

	@classmethod
	def detectInputFormat(cls, filename, format=""):
		if not format:
			ext = get_ext(filename)
			for key in Glossary.formatsExt.keys():
				if ext in Glossary.formatsExt[key]:
					format = key

		return format

	def clear(self) -> None:
		self._info = odict()

		self._data = []

		try:
			readers = self._readers
		except AttributeError:
			pass
		else:
			for reader in readers:
				try:
					reader.close()
				except Exception:
					log.exception("")
		self._readers = []

		self._iter = None
		self._entryFilters = []
		self._sortKey = None
		self._sortCacheSize = 1000

		self._filename = ""
		self._defaultDefiFormat = "m"
		self._progressbar = True

	def __init__(self, info: Optional[Dict[str, str]] = None, ui=Any) -> None:
		"""
		info:	OrderedDict instance, or None
				no need to copy OrderedDict instance,
				we will not reference to it
		"""
		self.clear()
		if info:
			if not isinstance(info, (dict, odict)):
				raise TypeError(
					"Glossary: `info` has invalid type"
					", dict or OrderedDict expected"
				)
			for key, value in info.items():
				self.setInfo(key, value)

		"""
		self._data is a list of tuples with length 2 or 3:
			(word, definition)
			(word, definition, defiFormat)
			where both word and definition can be a string, or list
				(containing word and alternates)

			defiFormat: format of the definition:
				"m": plain text
				"h": html
				"x": xdxf
		"""
		self.ui = ui

	def updateEntryFilters(self) -> None:
		from . import entry_filters as ef
		self._entryFilters = []
		pref = getattr(self.ui, "pref", {})

		self._entryFilters.append(ef.StripEntryFilter(self))
		self._entryFilters.append(ef.NonEmptyWordFilter(self))

		if pref.get("skipResources", False):
			self._entryFilters.append(ef.SkipDataEntryFilter(self))

		if pref.get("utf8Check", True):
			self._entryFilters.append(ef.FixUnicodeFilter(self))

		if pref.get("lower", True):
			self._entryFilters.append(ef.LowerWordFilter(self))

		if pref.get("remove_html_all", False):
			self._entryFilters.append(ef.RemoveHtmlTagsAll(self))
		elif pref.get("remove_html"):
			tags = pref.get("remove_html").split(",")
			self._entryFilters.append(ef.RemoveHtmlTags(self, tags))

		self._entryFilters.append(ef.LangEntryFilter(self))
		self._entryFilters.append(ef.CleanEntryFilter(self))
		self._entryFilters.append(ef.NonEmptyWordFilter(self))
		self._entryFilters.append(ef.NonEmptyDefiFilter(self))
		self._entryFilters.append(ef.RemoveEmptyAndDuplicateAltWords(self))

	def __str__(self) -> str:
		return "glossary.Glossary"

	def addEntryObj(self, entry: Entry) -> None:
		self._data.append(entry.getRaw())

	def newEntry(
		self,
		word: str,
		defi: str,
		defiFormat: str = "",
		byteProgress: Optional[Tuple[int, int]] = None,
	) -> Entry:
		"""
		create and return a new entry object
		"""
		if not defiFormat:
			defiFormat = self._defaultDefiFormat

		return Entry(
			word, defi,
			defiFormat=defiFormat,
			byteProgress=byteProgress,
		)

	def addEntry(self, word: str, defi: str, defiFormat: str = "") -> None:
		"""
		create and add a new entry object to glossary
		"""
		self.addEntryObj(self.newEntry(word, defi, defiFormat))

	def _loadedEntryGen(self) -> Iterator[BaseEntry]:
		wordCount = len(self._data)
		wcThreshold = wordCount // 200 + 1
		progressbar = self.ui and self._progressbar
		if progressbar:
			self.progressInit("Writing")
		for index, rawEntry in enumerate(self._data):
			if index & 0x7f == 0: # 0x3f, 0x7f, 0xff
				gc.collect()
			yield Entry.fromRaw(
				rawEntry,
				defaultDefiFormat=self._defaultDefiFormat
			)
			if progressbar and index % wcThreshold == 0:
				self.progress(index, wordCount)
		if progressbar:
			self.progressEnd()

	def _readersEntryGen(self) -> Iterator[BaseEntry]:
		for reader in self._readers:
			wordCount = 0
			progressbar = False
			if self.ui and self._progressbar:
				try:
					wordCount = len(reader)
				except Exception:
					log.exception("")
				progressbar = True
			if progressbar:
				self.progressInit("Converting")
			wcThreshold = wordCount // 200 + 1
			lastPos = 0
			try:
				for index, entry in enumerate(reader):
					yield entry
					if progressbar:
						if entry is None or wordCount > 0:
							if index % wcThreshold == 0:
								self.progress(index, wordCount)
							continue
						bp = entry.byteProgress()
						if bp and bp[0] > lastPos + 10000:
							self.progress(bp[0], bp[1], unit="bytes")
							lastPos = bp[0]
			finally:
				reader.close()
			if progressbar:
				self.progressEnd()

	def _applyEntryFiltersGen(
		self,
		gen: Iterator[BaseEntry],
	) -> Iterator[BaseEntry]:
		for entry in gen:
			if not entry:
				continue
			for entryFilter in self._entryFilters:
				entry = entryFilter.run(entry)
				if not entry:
					break
			else:
				yield entry

	def __iter__(self) -> Iterator[BaseEntry]:
		if self._iter is None:
			log.error(
				"Trying to iterate over a blank Glossary"
				", must call `glos.read` first"
			)
			return iter([])
		return self._iter

	def iterEntryBuckets(self, size: int) -> Iterator[BaseEntry]:
		"""
		iterate over buckets of entries, with size `size`
		For example:
			for bucket in glos.iterEntryBuckets(100):
				assert len(bucket) == 100
				for entry in bucket:
					print(entry.getWord())
				print("-----------------")
		"""
		bucket = []
		for entry in self:
			if len(bucket) >= size:
				yield bucket
				bucket = []
			bucket.append(entry)
		yield bucket

	def setDefaultDefiFormat(self, defiFormat: str) -> None:
		self._defaultDefiFormat = defiFormat

	def getDefaultDefiFormat(self) -> str:
		return self._defaultDefiFormat

	def __len__(self) -> int:
		return len(self._data) + sum(
			len(reader) for reader in self._readers
		)

	def infoKeys(self) -> List[str]:
		return list(self._info.keys())

	def getMostUsedDefiFormats(self, count: int = None) -> List[Tuple[str, int]]:
		return Counter([
			entry.getDefiFormat()
			for entry in self
		]).most_common(count)

	# def formatInfoKeys(self, format: str):# FIXME

	def iterInfo(self) -> Iterator[Tuple[str, str]]:
		return self._info.items()

	def getInfo(self, key: str) -> str:
		key = str(key) # FIXME: required?

		try:
			key = self.infoKeysAliasDict[key.lower()]
		except KeyError:
			pass

		return self._info.get(key, "")  # "" or None as default? FIXME

	def setInfo(self, key: str, value: str) -> None:
		#  FIXME
		origKey = key
		key = fixUtf8(key)
		value = fixUtf8(value)

		try:
			key = self.infoKeysAliasDict[key.lower()]
		except KeyError:
			pass

		if origKey != key:
			log.debug(f"setInfo: {origKey} -> {key}")

		self._info[key] = value

	def getExtraInfos(self, excludeKeys: List[str]) -> odict:
		"""
		excludeKeys: a list of (basic) info keys to be excluded
		returns an OrderedDict including the rest of info keys,
				with associated values
		"""
		excludeKeySet = set()
		for key in excludeKeys:
			excludeKeySet.add(key)
			try:
				excludeKeySet.add(self.infoKeysAliasDict[key.lower()])
			except KeyError:
				pass

		extra = odict()
		for key, value in self._info.items():
			if key in excludeKeySet:
				continue
			extra[key] = value

		return extra

	def getPref(self, name: str, default: Optional[str]) -> Optional[str]:
		if self.ui:
			return self.ui.pref.get(name, default)
		else:
			return default

	def newDataEntry(self, fname: str, data: bytes) -> DataEntry:
		inTmp = not self._readers
		return DataEntry(fname, data, inTmp)

	# ________________________________________________________________________#

	def read(
		self,
		filename: str,
		format: str = "",
		direct: bool = False,
		progressbar: bool = True,
		**options
	) -> bool:
		"""
		filename (str):	name/path of input file
		format (str):	name of input format,
						or "" to detect from file extension
		direct (bool):	enable direct mode
		"""
		filename = abspath(filename)

		# don't allow direct=False when there are readers
		# (read is called before with direct=True)
		if self._readers and not direct:
			raise ValueError(
				f"there are already {len(self._readers)} readers"
				f", you can not read with direct=False mode"
			)

		self.updateEntryFilters()
		###
		format = self.detectInputFormat(filename, format=format)
		if not format:
			log.error(f"Could not detect input format from file name: {filename!r}")
			return False

		validOptionKeys = self.formatsReadOptions[format]
		for key in list(options.keys()):
			if key not in validOptionKeys:
				log.error(
					f"Invalid read option {key!r} "
					f"given for {format} format"
				)
				del options[key]

		filenameNoExt, ext = splitext(filename)
		if not ext.lower() in self.formatsExt[format]:
			filenameNoExt = filename

		self._filename = filenameNoExt
		if not self.getInfo("name"):
			self.setInfo("name", split(filename)[1])
		self._progressbar = progressbar

		if format in self.readerClasses:
			Reader = self.readerClasses[format]
			reader = Reader(self)
			reader.open(filename, **options)
			if direct:
				self._readers.append(reader)
				log.info(
					f"Using Reader class from {format} plugin"
					f" for direct conversion without loading into memory"
				)
			else:
				self.loadReader(reader)
		else:
			if direct:
				log.debug(
					f"No `Reader` class found in {format} plugin"
					f", falling back to indirect mode"
				)
			result = self.readFunctions[format].__call__(
				self,
				filename,
				**options
			)
			# if not result:## FIXME
			#	return False

		self._updateIter()

		return True

	def loadReader(self, reader: Any) -> bool:
		"""
		iterates over `reader` object and loads the whole data into self._data
		must call `reader.open(filename)` before calling this function
		"""
		wordCount = 0
		progressbar = False
		if self.ui and self._progressbar:
			try:
				wordCount = len(reader)
			except Exception:
				log.exception("")
			progressbar = True
		if progressbar:
			self.progressInit("Reading")
		wcThreshold = wordCount // 200 + 1
		lastPos = 0
		try:
			for index, entry in enumerate(reader):
				if index & 0x7f == 0: # 0x3f, 0x7f, 0xff
					gc.collect()
				if entry:
					self.addEntryObj(entry)
				if progressbar:
					if entry is None or wordCount > 0:
						if index % wcThreshold == 0:
							self.progress(index, wordCount)
						continue
					bp = entry.byteProgress()
					if bp and bp[0] > lastPos + 20000:
						self.progress(bp[0], bp[1], unit="bytes")
						lastPos = bp[0]
		finally:
			reader.close()
		if progressbar:
			self.progressEnd()

		return True

	def _inactivateDirectMode(self) -> None:
		"""
		loads all of `self._readers` into `self._data`
		closes readers
		and sets self._readers to []
		"""
		for reader in self._readers:
			self.loadReader(reader)
		self._readers = []

	def _updateIter(self, sort: bool = False) -> None:
		"""
		updates self._iter
		depending on:
			1- Wheather or not direct mode is On (self._readers not empty)
				or Off (self._readers empty)
			2- Wheather sort is True, and if it is,
				checks for self._sortKey and self._sortCacheSize
		"""
		if self._readers:  # direct mode
			if sort:
				sortKey = self._sortKey
				cacheSize = self._sortCacheSize
				log.info(f"Stream sorting enabled, cache size: {cacheSize}")
				# only sort by main word, or list of words + alternates? FIXME
				gen = hsortStreamList(
					self._readers,
					cacheSize,
					key=Entry.getEntrySortKey(sortKey),
				)
			else:
				gen = self._readersEntryGen()
		else:
			gen = self._loadedEntryGen()

		self._iter = self._applyEntryFiltersGen(gen)

	def sortWords(
		self,
		key: Optional[Callable[[str], Any]] = None,
		cacheSize: int = 0,
	) -> None:
		# only sort by main word, or list of words + alternates? FIXME
		if self._readers:
			self._sortKey = key
			if cacheSize > 0:
				self._sortCacheSize = cacheSize  # FIXME
		else:
			self._data.sort(
				key=Entry.getRawEntrySortKey(key),
			)
		self._updateIter(sort=True)

	@classmethod
	def detectOutputFormat(
		cls,
		filename: str = "",
		format: str = "",
		inputFilename: str = "",
	) -> Optional[Tuple[str, str, str]]:
		"""
		returns (filename, format, archiveType) or None
		"""
		archiveType = ""
		if filename:
			ext = ""
			filenameNoExt, fext = splitext(filename)
			fext = fext.lower()
			if fext in (".gz", ".bz2", ".zip"):
				archiveType = fext[1:]
				filename = filenameNoExt
				fext = get_ext(filename)
			if not format:
				for fmt, extList in Glossary.formatsExt.items():
					for e in extList:
						if format == e[1:] or format == e:
							format = fmt
							ext = e
							break
					if format:
						break
				if not format:
					for fmt, extList in Glossary.formatsExt.items():
						if filename == fmt:
							if not inputFilename:
								log.error("inputFilename is empty")
							else:
								format = filename
								ext = extList[0]
								filename = inputFilename + ext
								break
						for e in extList:
							if not inputFilename:
								log.error("inputFilename is empty")
							elif filename == e[1:] or filename == e:
								format = fmt
								ext = e
								filename = inputFilename + ext
								break
						if format:
							break
				if not format:
					for fmt, extList in Glossary.formatsExt.items():
						if fext in extList:
							format = fmt
							ext = fext
			if not format:
				log.error("Unable to detect write format!")
				return

		else:  # filename is empty
			if not inputFilename:
				log.error(f"Invalid filename {filename!r}")
				return
			filename = inputFilename  # no extension
			if not format:
				log.error("No filename nor format is given for output file")
				return
			try:
				filename += Glossary.formatsExt[format][0]
			except KeyError:
				log.error("Invalid write format")
				return

		return filename, format, archiveType

	def write(
		self,
		filename: str,
		format: str,
		sort: Optional[bool] = None,
		sortKey: Optional[Callable[[str], Any]] = None,
		sortCacheSize: int = 1000,
		**options
	) -> Optional[str]:
		"""
		sort (bool):
			True (enable sorting),
			False (disable sorting),
			None (auto, get from UI)
		sortKey (callable or None):
			key function for sorting
			takes a word as argument, which is str or list (with alternates)

		returns absolute path of output file, or None if failed
		"""
		try:
			validOptionKeys = self.formatsWriteOptions[format]
		except KeyError:
			log.critical(f"No write support for {format!r} format")
			return
		for key in list(options.keys()):
			if key not in validOptionKeys:
				log.error(
					f"Invalid write option {key!r}"
					f" given for {format} format"
				)
				del options[key]

		plugin = self.plugins[format]
		sortOnWrite = plugin.sortOnWrite
		if sortOnWrite == ALWAYS:
			if sort is False:
				log.warning(
					f"Writing {format} requires sorting"
					f", ignoring user sort=False option"
				)
			if self._readers:
				log.warning(
					f"Writing to {format} format requires full sort"
					f", falling back to indirect mode"
				)
				self._inactivateDirectMode()
				log.info(f"Loaded {len(self._data)} entries")
			sort = True
		elif sortOnWrite == DEFAULT_YES:
			if sort is None:
				sort = True
		elif sortOnWrite == DEFAULT_NO:
			if sort is None:
				sort = False
		elif sortOnWrite == NEVER:
			if sort:
				log.warning(
					"Plugin prevents sorting before write" +
					", ignoring user sort=True option"
				)
			sort = False

		if sort:
			if sortKey is None:
				try:
					sortKey = plugin.sortKey
				except AttributeError:
					pass
				else:
					log.debug(
						"Using sort key function from %s plugin" % format
					)
			elif sortOnWrite == ALWAYS:
				try:
					sortKey = plugin.sortKey
				except AttributeError:
					pass
				else:
					log.warning(
						"Ignoring user-defined sort order, " +
						"and using key function from %s plugin" % format
					)
			self.sortWords(
				key=sortKey,
				cacheSize=sortCacheSize
			)
		else:
			self._updateIter(sort=False)

		filename = abspath(filename)
		log.info(f"Writing to file {filename!r}")
		try:
			if format in self.writerClasses:
				writer = self.writerClasses[format].__call__(self)
				writer.write(filename, **options)
			elif format in self.writeFunctions:
				self.writeFunctions[format].__call__(self, filename, **options)
			else:
				log.error(f"No write function or Writer class found for plugin {format}")
				return
		except Exception:
			log.exception("Exception while calling plugin\'s write function")
			return
		finally:
			self.clear()

		return filename

	def zipOutDir(self, filename: str):
		if isdir(filename):
			dirn, name = split(filename)
			with indir(filename):
				output, error = subprocess.Popen(
					["zip", "-r", f"../{name}.zip", ".", "-m"],
					stdout=subprocess.PIPE,
				).communicate()
				return error

		dirn, name = split(filename)
		with indir(dirn):
			output, error = subprocess.Popen(
				["zip", f"{filename}.zip", name, "-m"],
				stdout=subprocess.PIPE,
			).communicate()
			return error

	def archiveOutDir(self, filename: str, archiveType: str) -> str:
		"""
		filename is the existing file path
		archiveType is the archive extension (without dot): "gz", "bz2", "zip"
		"""
		try:
			os.remove("%s.%s" % (filename, archiveType))
		except OSError:
			pass
		if archiveType == "gz":
			output, error = subprocess.Popen(
				["gzip", filename],
				stdout=subprocess.PIPE,
			).communicate()
			if error:
				log.error(
					error + "\n" +
					"Failed to compress file \"%s\"" % filename
				)
		elif archiveType == "bz2":
			output, error = subprocess.Popen(
				["bzip2", filename],
				stdout=subprocess.PIPE,
			).communicate()
			if error:
				log.error(
					error + "\n" +
					"Failed to compress file \"%s\"" % filename
				)
		elif archiveType == "zip":
			error = self.zipOutDir(filename)
			if error:
				log.error(
					error + "\n" +
					"Failed to compress file \"%s\"" % filename
				)

		archiveFilename = "%s.%s" % (filename, archiveType)
		if isfile(archiveFilename):
			return archiveFilename
		else:
			return filename

	def convert(
		self,
		inputFilename: str,
		inputFormat: str = "",
		direct: Optional[bool] = None,
		progressbar: bool = True,
		outputFilename: str = "",
		outputFormat: str = "",
		sort: Optional[bool] = None,
		sortKey: Optional[Callable[[str], Any]] = None,
		sortCacheSize: int = 1000,
		readOptions: Optional[Dict[str, Any]] = None,
		writeOptions: Optional[Dict[str, Any]] = None,
	) -> Optional[str]:
		"""
		returns absolute path of output file, or None if failed
		"""
		if not readOptions:
			readOptions = {}
		if not writeOptions:
			writeOptions = {}

		if readOptions:
			log.info(f"readOptions = {readOptions}")
		if writeOptions:
			log.info(f"writeOptions = {writeOptions}")

		outputArgs = self.detectOutputFormat(
			filename=outputFilename,
			format=outputFormat,
			inputFilename=inputFilename,
		)
		if not outputArgs:
			log.error(f"Writing file {outputFilename!r} failed.")
			return
		outputFilename, outputFormat, archiveType = outputArgs

		if direct is None:
			if sort is not True:
				direct = True  # FIXME

		if isdir(outputFilename):
			log.error(f"Directory already exists: {outputFilename}")
			return

		tm0 = now()
		if not self.read(
			inputFilename,
			format=inputFormat,
			direct=direct,
			progressbar=progressbar,
			**readOptions
		):
			return
		log.info("")

		finalOutputFile = self.write(
			outputFilename,
			outputFormat,
			sort=sort,
			sortKey=sortKey,
			sortCacheSize=sortCacheSize,
			**writeOptions
		)
		log.info("")
		if not finalOutputFile:
			log.error(f"Writing file {outputFilename!r} failed.")
			return

		if archiveType:
			finalOutputFile = self.archiveOutDir(finalOutputFile, archiveType)

		log.info(f"Writing file {finalOutputFile!r} done.")
		log.info(f"Running time of convert: {now()-tm0:.1f} seconds")

		return finalOutputFile

	# ________________________________________________________________________#

	def writeTxt(
		self,
		sep1: str,
		sep2: str,
		filename: str = "",
		writeInfo: bool = True,
		rplList: Optional[List[Tuple[str, str]]] = None,
		ext: str = ".txt",
		head: str = "",
		iterEntries: Optional[Iterator[BaseEntry]] = None,
		entryFilterFunc: Optional[Callable[[BaseEntry], Optional[BaseEntry]]] = None,
		outInfoKeysAliasDict: Optional[Dict[str, str]] = None,
		encoding: str = "utf-8",
		newline: str = "\n",
		resources: bool = True,
	) -> bool:
		if rplList is None:
			rplList = []
		if not filename:
			filename = self._filename + ext
		if not outInfoKeysAliasDict:
			outInfoKeysAliasDict = {}

		fp = open(filename, "w", encoding=encoding, newline=newline)
		fp.write(head)
		if writeInfo:
			for key, desc in self._info.items():
				try:
					key = outInfoKeysAliasDict[key]
				except KeyError:
					pass
				for rpl in rplList:
					desc = desc.replace(rpl[0], rpl[1])
				fp.write("##" + key + sep1 + desc + sep2)
		fp.flush()

		myResDir = f"{filename}_res"
		if not isdir(myResDir):
			os.mkdir(myResDir)

		if not iterEntries:
			iterEntries = self

		for entry in iterEntries:
			if entry.isData():
				if resources:
					entry.save(myResDir)
				continue

			if entryFilterFunc:
				entry = entryFilterFunc(entry)
				if not entry:
					continue
			word = entry.getWord()
			defi = entry.getDefi()
			if word.startswith("#"):  # FIXME
				continue
			# if self.getPref("enable_alts", True):  # FIXME

			for rpl in rplList:
				defi = defi.replace(rpl[0], rpl[1])
			fp.write(word + sep1 + defi + sep2)
		fp.close()
		if not os.listdir(myResDir):
			os.rmdir(myResDir)
		return True

	def writeTabfile(self, filename: str = "", **kwargs) -> None:
		self.writeTxt(
			"\t",
			"\n",
			filename=filename,
			rplList=(
				("\\", "\\\\"),
				("\n", "\\n"),
				("\t", "\\t"),
			),
			ext=".txt",
			**kwargs
		)

	def writeDict(self, filename: str = "", writeInfo: bool = False) -> None:
		# Used in "/usr/share/dict/" for some dictionarys such as "ding"
		self.writeTxt(
			" :: ",
			"\n",
			filename,
			writeInfo,
			(
				("\n", "\\n"),
			),
			".dict",
		)

	def iterSqlLines(
		self,
		filename: str = "",
		infoKeys: Optional[List] = None,
		addExtraInfo: bool = True,
		newline: str = "\\n",
		transaction: bool = False,
	) -> Iterator[str]:
		newline = "<br>"
		infoDefLine = "CREATE TABLE dbinfo ("
		infoValues = []

		if not infoKeys:
			infoKeys = [
				"dbname",
				"author",
				"version",
				"direction",
				"origLang",
				"destLang",
				"license",
				"category",
				"description",
			]

		for key in infoKeys:
			value = self.getInfo(key)
			value = value\
				.replace("\'", "\'\'")\
				.replace("\x00", "")\
				.replace("\r", "")\
				.replace("\n", newline)
			infoValues.append(f"\'{value}\'")
			infoDefLine += f"{key} char({len(value)}), "

		infoDefLine = infoDefLine[:-2] + ");"
		yield infoDefLine

		if addExtraInfo:
			yield (
				"CREATE TABLE dbinfo_extra (" +
				"\'id\' INTEGER PRIMARY KEY NOT NULL, " +
				"\'name\' TEXT UNIQUE, \'value\' TEXT);"
			)

		yield (
			"CREATE TABLE word (\'id\' INTEGER PRIMARY KEY NOT NULL, " +
			"\'w\' TEXT, \'m\' TEXT);"
		)

		if transaction:
			yield "BEGIN TRANSACTION;"
		yield f"INSERT INTO dbinfo VALUES({','.join(infoValues)});"

		if addExtraInfo:
			extraInfo = self.getExtraInfos(infoKeys)
			for index, (key, value) in enumerate(extraInfo.items()):
				key = key.replace("\'", "\'\'")
				value = value.replace("\'", "\'\'")
				yield (
					f"INSERT INTO dbinfo_extra VALUES({index+1}, "
					f"\'{key}\', \'{value}\');"
				)

		for i, entry in enumerate(self):
			if entry.isData():
				# FIXME
				continue
			word = entry.getWord()
			defi = entry.getDefi()
			word = word.replace("\'", "\'\'")\
				.replace("\r", "").replace("\n", newline)
			defi = defi.replace("\'", "\'\'")\
				.replace("\r", "").replace("\n", newline)
			yield f"INSERT INTO word VALUES({i+1}, \'{word}\', \'{defi}\');"
		if transaction:
			yield "END TRANSACTION;"
		yield "CREATE INDEX ix_word_w ON word(w COLLATE NOCASE);"

	# ________________________________________________________________________#

	def takeOutputWords(self, minWordLen: int = 3) -> List[str]:
		# fr"[\w]{{{minWordLen},}}"
		wordPattern = re.compile(r"[\w]{%d,}" % minWordLen, re.U)
		words = set()
		progressbar, self._progressbar = self._progressbar, False
		for entry in self:
			words.update(re.findall(
				wordPattern,
				entry.getDefi(),
			))
		self._progressbar = progressbar
		return sorted(words)

	# ________________________________________________________________________#

	def progressInit(self, *args) -> None:
		if self.ui:
			self.ui.progressInit(*args)

	def progress(self, pos: int, total: int, unit: str = "entries") -> None:
		if not self.ui:
			return
		self.ui.progress(
			min(pos + 1, total) / total,
			f"{pos:d} / {total:d} {unit}",
		)

	def progressEnd(self) -> None:
		if self.ui:
			self.ui.progressEnd()

	# ________________________________________________________________________#

	def searchWordInDef(
		self,
		st: str,
		matchWord: bool = True,
		sepChars: str = ".,،",
		maxNum: int = 100,
		minRel: float = 0.0,
		minWordLen: int = 3,
		includeDefs: bool = False,
		showRel: str = "Percent", # "Percent" | "Percent At First" | ""
	) -> List[str]:
		# searches word "st" in definitions of the glossary
		splitPattern = re.compile(
			"|".join([re.escape(x) for x in sepChars]),
			re.U,
		)
		wordPattern = re.compile(r"[\w]{%d,}" % minWordLen, re.U)
		outRel = []
		for item in self._data:
			words, defi = item[:2]
			if isinstance(words, str):
				words = [words]
			if isinstance(defi, list):
				defi = "\n".join(defi)
			if st not in defi:
				continue
			for word in words:
				rel = 0  # relation value of word (0 <= rel <= 1)
				for part in re.split(splitPattern, defi):
					if not part:
						continue
					if matchWord:
						partWords = re.findall(
							wordPattern,
							part,
						)
						if not partWords:
							continue
						rel = max(
							rel,
							partWords.count(st) / len(partWords)
						)
					else:
						rel = max(
							rel,
							part.count(st) * len(st) / len(part)
						)
				if rel <= minRel:
					continue
				if includeDefs:
					outRel.append((word, rel, defi))
				else:
					outRel.append((word, rel))
		outRel.sort(
			key=lambda x: x[1],
			reverse=True,
		)
		n = len(outRel)
		if n > maxNum > 0:
			outRel = outRel[:maxNum]
			n = maxNum
		num = 0
		out = []
		if includeDefs:
			for j in range(n):
				numP = num
				w, num, m = outRel[j]
				m = m.replace("\n", "\\n").replace("\t", "\\t")
				onePer = int(1.0 / num)
				if onePer == 1.0:
					out.append(f"{w}\\n{m}")
				elif showRel == "Percent":
					out.append(f"{w}(%{100*num})\\n{m}")
				elif showRel == "Percent At First":
					if num == numP:
						out.append(f"{w}\\n{m}")
					else:
						out.append(f"{w}(%{100*num})\\n{m}")
				else:
					out.append(f"{w}\\n{m}")
			return out
		for j in range(n):
			numP = num
			w, num = outRel[j]
			onePer = int(1.0 / num)
			if onePer == 1.0:
				out.append(w)
			elif showRel == "Percent":
				out.append(f"{w}(%{100*num})")
			elif showRel == "Percent At First":
				if num == numP:
					out.append(w)
				else:
					out.append(f"{w}(%{100*num})")
			else:
				out.append(w)
		return out

	def reverse(
		self,
		savePath: str = "",
		words: Optional[List[str]] = None,
		includeDefs: bool = False,
		reportStep: int = 300,
		saveStep: int = 1000,  # set this to zero to disable auto saving
		**kwargs
	) -> Iterator[int]:
		"""
		This is a generator
		Usage:
			for wordIndex in glos.reverse(...):
				pass

		Inside the `for` loop, you can pause by waiting (for input or a flag)
			or stop by breaking

		Potential keyword arguments:
			words = None ## None, or list
			reportStep = 300
			saveStep = 1000
			savePath = ""
			matchWord = True
			sepChars = ".,،"
			maxNum = 100
			minRel = 0.0
			minWordLen = 3
			includeDefs = False
			showRel = "None"
				allowed values: "None", "Percent", "Percent At First"
		"""
		if not savePath:
			savePath = self.getInfo("name") + ".txt"

		if saveStep < 2:
			raise ValueError("saveStep must be more than 1")

		ui = self.ui

		if words:
			words = list(words)
		else:
			words = self.takeOutputWords()

		wordCount = len(words)
		log.info(
			f"Reversing to file {savePath!r}"
			f", number of words: {wordCount}"
		)
		self.progressInit("Reversing")
		wcThreshold = wordCount // 200 + 1
		with open(savePath, "w") as saveFile:
			for wordI in range(wordCount):
				word = words[wordI]
				if wordI % wcThreshold == 0:
					self.progress(wordI, wordCount)

				if wordI % saveStep == 0 and wordI > 0:
					saveFile.flush()
				result = self.searchWordInDef(
					word,
					includeDefs=includeDefs,
					**kwargs
				)
				if result:
					try:
						if includeDefs:
							defi = "\\n\\n".join(result)
						else:
							defi = ", ".join(result) + "."
					except Exception:
						log.exception("")
						log.debug(f"result = {result}")
						return
					saveFile.write(f"{word}\t{defi}\n")
				yield wordI

		self.progressEnd()
		yield wordCount

	@classmethod
	def init(cls):
		cls.readFormats = []
		cls.writeFormats = []
		cls.readExt = []
		cls.writeExt = []
		cls.readDesc = []
		cls.writeDesc = []
		cls.loadPlugins(join(dirname(__file__), "plugins"))
		cls.loadPlugins(userPluginsDir)


Glossary.init()



# coding=utf-8
# based on https://github.com/zkoch/CEDICT_Parser

TONES = {
	"a1":"ā", "a2":"á", "a3":"ǎ", "a4":"à",
	"e1":"ē", "e2":"é", "e3":"ě", "e4":"è",
	"i1":"ī", "i2":"í", "i3":"ǐ", "i4":"ì",
	"o1":"ō", "o2":"ó", "o3":"ǒ", "o4":"ò",
	"u1":"ū", "u2":"ú", "u3":"ǔ", "u4":"ù",
	"v1":"ǖ", "v2":"ǘ", "v3":"ǚ", "v4":"ǜ",
}
# using v for the umlauted u

VOWELS = ("a", "e", "o", "iu", "ui", "i", "u", "v")

def convert(word):
	tone = word[-1]
	pinyin = word[0:-1].lower()
	result = pinyin

	if tone == "5":
		return pinyin, tone
	elif tone not in ("1", "2", "3", "4"):
		return word, ""

	for vowel in VOWELS:
		if vowel in pinyin:
			vowel1 = vowel[-1]
			result = pinyin.replace(vowel1, TONES[vowel1 + tone])
			break

	return result, tone




# -*- coding: utf-8 -*-
# The MIT License (MIT)

# Copyright (C) 2012-2016 Alberto Pettarin (alberto@albertopettarin.it)
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from formats_common import *
from pyglossary.ebook_base import *

enable = True
format = "Epub2"
description = "EPUB-2 E-Book"
extensions = [".epub",]
sortOnWrite = ALWAYS

optionsProp = {
	"group_by_prefix_length": IntOption(),
	# "group_by_prefix_merge_min_size": IntOption(),
	# "group_by_prefix_merge_across_first": BoolOption(),
	"compress": BoolOption(),
	"keep": BoolOption(),
	"include_index_page": BoolOption(),
	"apply_css": StrOption(comment="path to css file"),
	"cover_path": StrOption(comment="path to cover file"),
}

class Writer(EbookWriter):
	ebook_format = format
	## MIMETYPE_CONTENTS, CONTAINER_XML_CONTENTS, NCX_TEMPLATE, NCX_NAVPOINT_TEMPLATE are only in Epub
	MIMETYPE_CONTENTS = "application/epub+zip"
	CONTAINER_XML_CONTENTS = """<?xml version="1.0" encoding="UTF-8" ?>
<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
   <rootfiles>
	  <rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
   </rootfiles>
</container>"""

	NCX_TEMPLATE = """<?xml version="1.0" encoding="utf-8" ?>
<!DOCTYPE ncx PUBLIC "-//NISO//DTD ncx 2005-1//EN" "http://www.daisy.org/z3986/2005/ncx-2005-1.dtd">
<ncx xmlns="http://www.daisy.org/z3986/2005/ncx/" version="2005-1">
 <head>
  <meta name="dtb:uid" content="{identifier}" />
  <meta name="dtb:depth" content="1" />
  <meta name="dtb:totalPageCount" content="0" />
  <meta name="dtb:maxPageNumber" content="0" />
 </head>
 <docTitle>
  <text>{title}</text>
 </docTitle>
 <navMap>
{ncx_items}
 </navMap>
</ncx>"""

	NCX_NAVPOINT_TEMPLATE = """ <navPoint id="n{index:06d}" playOrder="{index:d}">
  <navLabel>
   <text>{text}</text>
  </navLabel>
  <content src="{src}" />
 </navPoint>"""

	CSS_CONTENTS = """@charset "UTF-8";
body {
  margin: 10px 25px 10px 25px;
}  
h1 {
  font-size: 200%;
}
h2 {
  font-size: 150%;
}
p {
  margin-left: 0em;
  margin-right: 0em;
  margin-top: 0em;
  margin-bottom: 0em;
  line-height: 2em;
  text-align: justify;
}
a, a:focus, a:active, a:visited {
  color: black;
  text-decoration: none;
}
body.indexPage {}
h1.indexTitle {}
p.indexGroups {
  font-size: 150%;
}
span.indexGroup {}
body.groupPage {}
h1.groupTitle {}
div.groupNavigation {}
span.groupHeadword {}
div.groupEntry {
  margin-top: 0;
  margin-bottom: 1em;
}
h2.groupHeadword {
  margin-left: 5%;
}
p.groupDefinition {
  margin-left: 10%;
  margin-right: 10%;
}
"""


	GROUP_XHTML_TEMPLATE = """<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>{title}</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body id="groupPage" class="groupPage">
  <h1 class="groupTitle">{group_title}</h1>
  <div class="groupNavigation">
   <a href="{previous_link}">[ Previous ]</a>
{index_link}
   <a href="{next_link}">[ Next ]</a>
  </div>
{group_contents}
 </body>
</html>"""
	GROUP_XHTML_INDEX_LINK = "   <a href=\"index.xhtml\">[ Index ]</a>"

	GROUP_XHTML_WORD_TEMPLATE = "   <span class=\"groupHeadword\">{headword}</span>"

	GROUP_XHTML_WORD_JOINER = " &#8226;\n"

	GROUP_XHTML_WORD_DEFINITION_TEMPLATE = """  <div class="groupEntry">
   <h2 class="groupHeadword">{headword}</h2>
   <p class="groupDefinition">{definition}</p>
  </div>"""

	OPF_TEMPLATE = """<?xml version="1.0" encoding="utf-8" ?>
<package xmlns="http://www.idpf.org/2007/opf" version="2.0" unique-identifier="uid">
	<metadata xmlns:opf="http://www.idpf.org/2007/opf" xmlns:dc="http://purl.org/dc/elements/1.1/">
		<dc:identifier id="uid" opf:scheme="uuid">{identifier}</dc:identifier>
		<dc:language>{sourceLang}</dc:language>
		<dc:title>{title}</dc:title>
		<dc:creator opf:role="aut">{creator}</dc:creator>
		<dc:rights>{copyright}</dc:rights>
		<dc:date opf:event="creation">{creationDate}</dc:date>
		{cover}
	</metadata>
 <manifest>
{manifest}
 </manifest>
 <spine toc="toc.ncx">
{spine}
 </spine>
</package>"""


	COVER_TEMPLATE = "<meta name=\"cover\" content=\"{cover}\" />"

	def __init__(self, glos):
		import uuid
		EbookWriter.__init__(
			self,
			glos,
		)
		glos.setInfo(
			"uuid",
			str(uuid.uuid4()).replace("-", ""),
		)


	def write_ncx(self, group_labels, include_index_page):
		"""
			write_ncx
			only for epub
		"""
		ncx_items = []
		index = 1
		if include_index_page:
			ncx_items.append(self.NCX_NAVPOINT_TEMPLATE.format(
				index=index,
				text="Index",
				src="index.xhtml",
			))
			index += 1
		for group_label in group_labels:
			ncx_items.append(self.NCX_NAVPOINT_TEMPLATE.format(
				index = index,
				text = group_label,
				src = self.get_group_xhtml_file_name_from_index(index),
			))
			index += 1
		ncx_items_unicode = "\n".join(ncx_items)
		ncx_contents = self.NCX_TEMPLATE.format(
			identifier = self.glos.getInfo("uuid"),
			title = self.glos.getInfo("title"),
			ncx_items = ncx_items_unicode,
		)
		self.add_file_manifest(
			"OEBPS/toc.ncx",
			"toc.ncx",
			ncx_contents,
			"application/x-dtbncx+xml",
		)

	# inherts write from EbookWriter




# -*- coding: utf-8 -*-
# The MIT License (MIT)

# Copyright (C) 2012-2016 Alberto Pettarin (alberto@albertopettarin.it)
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
log = logging.getLogger("root")


from itertools import groupby
import os
from os.path import join
import zipfile
import tempfile
from datetime import datetime
import shutil

from pyglossary.text_utils import toStr, toBytes
from pyglossary.os_utils import indir


def get_prefix(word, length):
	"""
	Return the prefix for the given word,
	of length length.

	:param word: the word string
	:type  word: unicode
	:param length: prefix length
	:type  length: int
	:rtype: unicode
	"""
	if not word:
		return None
	word = toStr(word)
	if "Z" < word[0] < "a":
		return "SPECIAL"
	return word[:length] ## return a unicode? FIXME


class EbookWriter(object):
	"""
	A class representing a generic ebook containing a dictionary.

	It can be used to output a MOBI or an EPUB 2 container.

	The ebook must have an OPF, and one or more group XHTML files.

	Optionally, it can have a cover image, an NCX TOC, an index XHTML file.

	The actual file templates are provided by the caller.
	"""
	ebook_format = ""

	CSS_CONTENTS = ""
	GROUP_XHTML_TEMPLATE = ""
	GROUP_XHTML_INDEX_LINK = ""
	
	GROUP_XHTML_WORD_TEMPLATE = ""
	GROUP_XHTML_WORD_JOINER = ""
	GROUP_XHTML_WORD_DEFINITION_TEMPLATE = ""
	GROUP_XHTML_WORD_DEFINITION_JOINER = "\n"
	
	MIMETYPE_CONTENTS = ""
	CONTAINER_XML_CONTENTS = ""
	
	GROUP_START_INDEX = 2

	COVER_TEMPLATE = "{cover}"

	INDEX_XHTML_TEMPLATE = """<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>{title}</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body class="indexPage">
  <h1 class="indexTitle">{indexTitle}</h1>
  <p class="indexGroupss">
{links}
  </p>
 </body>
</html>"""
	INDEX_XHTML_LINK_TEMPLATE = "   <span class=\"indexGroup\"><a href=\"{ref}\">{label}</a></span>"

	INDEX_XHTML_LINK_JOINER = " &#8226;\n"


	OPF_MANIFEST_ITEM_TEMPLATE = "  <item href=\"{ref}\" id=\"{id}\" media-type=\"{mediaType}\" />"

	OPF_SPINE_ITEMREF_TEMPLATE = "  <itemref idref=\"{id}\" />"


	def get_opf_contents(self, manifest_contents, spine_contents):
		raise NotImplementedError

	def __init__(
		self,
		glos,
		escape_strings=False,
		ignore_synonyms=False,
		flatten_synonyms=False,
	):
		self.glos = glos

		self._escape_strings = escape_strings
		self._ignore_synonyms = ignore_synonyms
		self._flatten_synonyms = flatten_synonyms

		#"bookeen_collation_function": None,## bookeen format
		#"bookeen_install_file": False,## bookeen format
		#"marisa_bin_path": None,## kobo format
		#"marisa_index_size": 1000000,## kobo format
		#"sd_ignore_sametypesequence": False,## stardict format
		#"sd_no_dictzip": False,## stardict format
		
		#"group_by_prefix_merge_across_first": False,
		#"group_by_prefix_merge_min_size": 0,

		self.tmpDir = None
		self.cover = None
		self.files = []
		self.manifest_files = []
		self.groups = []

	def close(self):
		pass

	myOpen = lambda self, fname, mode: open(join(self.tmpDir, fname), mode)

	def add_file(self, relative_path, contents, mode=None):
		if mode is None:
			mode = zipfile.ZIP_DEFLATED
		file_path = os.path.join(self.tmpDir, relative_path)
		file_obj = self.myOpen(file_path, "wb")
		contents = toBytes(contents)
		file_obj.write(contents)
		file_obj.close()
		self.files.append({
			"path": relative_path,
			"mode": mode,
		})

	def write_cover(self, cover_path):
		basename = os.path.basename(cover_path)
		cover_obj = self.myOpen(cover_path, "rb")
		cover = cover_obj.read()
		cover_obj.close()
		b = basename.lower()
		mimetype = "image/jpeg"
		if b.endswith(".png"):
			mimetype = "image/png"
		elif b.endswith(".gif"):
			mimetype = "image/gif"
		self.add_file_manifest("OEBPS/%s" % basename, basename, cover, mimetype)
		self.cover = basename

	def write_css(self, custom_css_path_absolute):
		css = self.CSS_CONTENTS
		if custom_css_path_absolute is not None:
			try:
				css_obj = self.myOpen(custom_css_path_absolute, "rb")
				css = css_obj.read()
				css_obj.close()
			except:
				pass
		self.add_file_manifest("OEBPS/style.css", "style.css", css, "text/css")

	def add_file_manifest(self, relative_path, id, contents, mimetype):
		self.add_file(relative_path, contents)
		self.manifest_files.append({
			"path": relative_path,
			"id": id, "mimetype": mimetype,
		})

	def get_group_xhtml_file_name_from_index(self, index):
		if index < self.GROUP_START_INDEX:
		## or index >= len(self.groups) + self.GROUP_START_INDEX:
		## number of groups are not known## FIXME
		## so we can not say if the current group is the last or not
			return "#groupPage"
		return "g%06d.xhtml" % index

	def add_group(self, key, entries):
		self.groups.append({"key": key, "entries": entries})

	def write_groups(self, group_by_prefix_length, include_index_page):
		# TODO: rtl=False option
		# TODO: handle alternates better (now shows word1|word2... in title)

		group_labels = []
		
		self.glos.sortWords()
		for group_i, (group_prefix, group_entry_iter) in enumerate(groupby(
			self.glos,
			lambda tmpEntry: get_prefix(
				tmpEntry.getWord(),
				group_by_prefix_length,
			),
		)):
			index = group_i + self.GROUP_START_INDEX
			first_word = ""
			last_word = ""
			group_contents = []
			for entry in group_entry_iter:
				if entry.isData():
					continue
				word = entry.getWord()
				defi = entry.getDefi()
				if not first_word:
					first_word = word
				last_word = word
				group_contents.append(self.GROUP_XHTML_WORD_DEFINITION_TEMPLATE.format(
					headword=self.escape_if_needed(word),
					definition=self.escape_if_needed(defi),
				))

			group_label = group_prefix
			if group_prefix != "SPECIAL":
				group_label = "%s&#8211;%s" % (first_word, last_word)
			group_labels.append(group_label)

			previous_link = self.get_group_xhtml_file_name_from_index(index - 1)
			next_link = self.get_group_xhtml_file_name_from_index(index + 1)

			group_contents = self.GROUP_XHTML_WORD_DEFINITION_JOINER.join(group_contents)
			group_contents = self.GROUP_XHTML_TEMPLATE.format(
				title=group_label,
				group_title=group_label,
				previous_link=previous_link,
				index_link=self.GROUP_XHTML_INDEX_LINK if include_index_page else "",
				next_link=next_link,
				group_contents=group_contents,
			)
			
			group_xhtml_path = self.get_group_xhtml_file_name_from_index(index)

			self.add_file_manifest(
				"OEBPS/%s" % group_xhtml_path,
				group_xhtml_path,
				group_contents,
				"application/xhtml+xml",
			)
		
		return group_labels

	def escape_if_needed(self, string):
		if self._escape_strings:
			string = string.replace("&", "&amp;")\
				.replace('"', "&quot;")\
				.replace("'", "&apos;")\
				.replace(">", "&gt;")\
				.replace("<", "&lt;")
		return string


	def write_index(self, group_labels):
		"""
			group_labels: a list of labels
		"""
		links = []
		for label_i, label in enumerate(group_labels):
			links.append(self.INDEX_XHTML_LINK_TEMPLATE.format(
				ref = self.get_group_xhtml_file_name_from_index(
					self.GROUP_START_INDEX + label_i
				),
				label = label,
			))
		links = self.INDEX_XHTML_LINK_JOINER.join(links)
		title = self.glos.getInfo("title")
		contents = self.INDEX_XHTML_TEMPLATE.format(
			title = title,
			indexTitle = title,
			links = links,
		)
		self.add_file_manifest(
			"OEBPS/index.xhtml",
			"index.xhtml",
			contents,
			"application/xhtml+xml",
		)

	def get_opf_contents(self, manifest_contents, spine_contents):
		cover = ""
		if self.cover:
			cover = self.COVER_TEMPLATE.format(cover=self.cover)

		creationDate = datetime.now().strftime("%Y-%m-%d")
		
		return self.OPF_TEMPLATE.format(
			identifier = self.glos.getInfo("uuid"),
			sourceLang = self.glos.getInfo("sourceLang"),
			targetLang = self.glos.getInfo("sourceLang"),
			title = self.glos.getInfo("title"),
			creator = self.glos.getInfo("author"),
			copyright = self.glos.getInfo("copyright"),
			creationDate = creationDate,
			cover = cover,
			manifest = manifest_contents,
			spine = spine_contents,
		)

	def write_opf(self):
		manifest_lines = []
		spine_lines = []
		for mi in self.manifest_files:
			manifest_lines.append(self.OPF_MANIFEST_ITEM_TEMPLATE.format(
				ref=mi["id"],
				id=mi["id"],
				mediaType=mi["mimetype"]
			))
			if mi["mimetype"] == "application/xhtml+xml":
				spine_lines.append(self.OPF_SPINE_ITEMREF_TEMPLATE.format(
					id=mi["id"],
				))

		manifest_contents = "\n".join(manifest_lines)
		spine_contents = "\n".join(spine_lines)
		opf_contents = self.get_opf_contents(
			manifest_contents,
			spine_contents,
		)

		self.add_file("OEBPS/content.opf", opf_contents)

	def write_ncx(self, group_labels, include_index_page):
		"""
			write_ncx
			only for epub
		"""
		pass

	def write(
		self,
		filename,
		keep=False,
		group_by_prefix_length=2,
		include_index_page=False,
		compress=True,
		apply_css="", # path to css file, or ""
		cover_path="", # path to cover file, or ""
	):
		self.tmpDir = tempfile.mkdtemp()
		with indir(self.tmpDir):
			if cover_path:
				cover_path = os.path.abspath(cover_path)

			if apply_css:
				apply_css = os.path.abspath(apply_css)

			os.makedirs("META-INF")
			os.makedirs("OEBPS")

			if self.MIMETYPE_CONTENTS:
				self.add_file("mimetype", self.MIMETYPE_CONTENTS, mode=zipfile.ZIP_STORED)
			if self.CONTAINER_XML_CONTENTS:
				self.add_file("META-INF/container.xml", self.CONTAINER_XML_CONTENTS)

			if cover_path:
				try:
					self.write_cover(cover_path)
				except:
					log.exception("")

			self.write_css(apply_css)

			group_labels = self.write_groups(
				group_by_prefix_length,
				include_index_page,
			)

			if include_index_page:
				self.write_index()

			self.write_ncx(group_labels, include_index_page)

			self.write_opf()

			if compress:
				zipFp = zipfile.ZipFile(
					filename,
					"w",
					compression=zipfile.ZIP_DEFLATED,
				)
				for fileDict in self.files:
					zipFp.write(
						fileDict["path"],
						compress_type=fileDict["mode"],
					)
				zipFp.close()
				if not keep:
					shutil.rmtree(self.tmpDir)
			else:
				if keep:
					shutil.copytree(self.tmpDir, filename)
				else:
					shutil.move(self.tmpDir, filename)







# -*- coding: utf-8 -*-

import logging
from typing import Tuple, List, Dict, Optional, Any


log = logging.getLogger("root")


class Option(object):
	def __init__(
		self,
		typ: str,
		customValue: bool = False,
		values: Optional[List[str]] = None,
		comment: str = "",
		disabled: bool = False,
	):
		if values is None:
			customValue = True # otherwise there would not be any valid value
		self.typ = typ
		self.values = values
		self.customValue = customValue
		self.comment = comment
		self.disabled = disabled

	def evaluate(self, raw: str) -> Tuple[Any, bool]:
		"returns (value, isValid)"
		return raw, True

	def validate(self, value):
		if not self.customValue:
			if not self.values:
				log.error("invalid option: customValue=%r, values=%r" % (
					self.customValue,
					self.values,
				))
				return False
			return value in self.values
		if value is None:
			return self.typ in ("dict", "list")
		valueType = type(value).__name__
		return self.typ == valueType

	def validateRaw(self, raw: str) -> bool:
		"returns isValid"
		value, isValid = self.evaluate(raw)
		if not isValid:
			return False
		if not self.validate(value):
			return False
		return True

	def groupValues(self) -> Optional[Dict[str, Any]]:
		return None


class BoolOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(
			self,
			"bool",
			customValue=False,
			values=[False, True],
			**kwargs,
		)

	def evaluate(self, raw: str) -> Tuple[Optional[bool], bool]:
		if raw.lower() in ("yes", "true", "1"):
			return True, True
		if raw.lower() in ("no", "false", "0"):
			return False, True
		return None, False # not valid


class StrOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(self, "str", **kwargs)

	def validate(self, value):
		if not self.customValue:
			if not self.values:
				log.error("invalid option: customValue=%r, values=%r" % (
					self.customValue,
					self.values,
				))
				return False
			return value in self.values
		return type(value).__name__ == "str"

	def groupValues(self) -> Optional[Dict[str, Any]]:
		return None


class IntOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(self, "int", **kwargs)

	def evaluate(self, raw: str) -> Tuple[Optional[int], bool]:
		"returns (value, isValid)"
		try:
			value = int(raw)
		except ValueError:
			return None, False
		else:
			return value, True


class FloatOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(self, "float", **kwargs)

	def evaluate(self, raw: float) -> Tuple[Optional[float], bool]:
		"returns (value, isValid)"
		try:
			value = float(raw)
		except ValueError:
			return None, False
		else:
			return value, True


class DictOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(
			self,
			"dict",
			customValue=True,
			**kwargs,
		)

	def evaluate(self, raw: str) -> Tuple[Optional[Dict], bool]:
		import ast
		if raw == "":
			return None, True # valid
		try:
			value = ast.literal_eval(raw)
		except SyntaxError:
			return None, False # not valid
		if type(value).__name__ != "dict":
			return None, False # not valid
		return value, True # valid


class ListOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(
			self,
			"list",
			customValue=True,
			**kwargs,
		)

	def evaluate(self, raw: str) -> Tuple[Optional[List], bool]:
		import ast
		if raw == "":
			return None, True # valid
		try:
			value = ast.literal_eval(raw)
		except SyntaxError:
			return None, False # not valid
		if type(value).__name__ != "list":
			return None, False # not valid
		return value, True # valid


class EncodingOption(Option):
	def __init__(self, customValue=True, values=None, **kwargs):
		if values is None:
			values = [
				"utf-8",
				"utf-16",
				"windows-1250",
				"windows-1251",
				"windows-1252",
				"windows-1253",
				"windows-1254",
				"windows-1255",
				"windows-1256",
				"windows-1257",
				"windows-1258",
				"mac_cyrillic",
				"mac_greek",
				"mac_iceland",
				"mac_latin2",
				"mac_roman",
				"mac_turkish",
				"cyrillic",
				"arabic",
				"greek",
				"hebrew",
				"latin2",
				"latin3",
				"latin4",
				"latin5",
				"latin6",
			]
		Option.__init__(
			self,
			"str",
			customValue=customValue,
			values=values,
			**kwargs
		)

	def groupValues(self) -> Optional[Dict[str, Any]]:
		import re
		from collections import OrderedDict
		groups = OrderedDict() # type: Dict[str, List[str]]
		others = [] # type: List[str]
		for value in self.values:
			cats = re.findall("^[a-z]+", value)
			if not cats:
				others.append(value)
				continue
			cat = cats[0]
			if len(cat) == len(value):
				others.append(value)
				continue
			if not cat in groups:
				groups[cat] = []
			groups[cat].append(value)
		if others:
			groups["other"] = others
		return groups


class NewlineOption(Option):
	def __init__(self, customValue=True, values=None, **kwargs):
		if values is None:
			values = [
				"\r\n",
				"\n",
				"\r",
			]
		Option.__init__(
			self,
			"str",
			customValue=customValue,
			values=values,
			**kwargs
		)


class HtmlColorOption(Option):
	def __init__(self, **kwargs):
		Option.__init__(self, "str", customValue=True, **kwargs)
		# FIXME: use a specific type?



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = "Sql"
description = "SQL"
extensions = [".sql"]
optionsProp = {
	"encoding": EncodingOption(),
}
depends = {}


def write(
	glos: GlossaryType,
	filename: str,
	encoding: str = "utf-8",
):
	with open(filename, "w", encoding=encoding) as fp:
		for line in glos.iterSqlLines(
			transaction=False,
		):
			fp.write(line + "\n")



# -*- coding: utf-8 -*-
#
# Copyright © 2013 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from formats_common import *
import csv
from pyglossary.file_utils import fileCountLines


enable = True
format = "Csv"
description = "CSV"
extensions = [".csv"]
optionsProp = {
	"encoding": EncodingOption(),
	"resources": BoolOption(),
}
depends = {}
supportsAlternates = True


class Reader(object):
	def __init__(self, glos: GlossaryType):
		self._glos = glos
		self.clear()

	def clear(self) -> None:
		self._filename = ""
		self._file = None
		self._leadingLinesCount = 0
		self._wordCount = None
		self._pos = -1
		self._csvReader = None
		self._resDir = ""
		self._resFileNames = []

	def open(self, filename: str, encoding: str = "utf-8") -> None:
		self._filename = filename
		self._file = open(filename, "r", encoding=encoding)
		self._csvReader = csv.reader(
			self._file,
			dialect="excel",
		)
		self._resDir = filename + "_res"
		if isdir(self._resDir):
			self._resFileNames = os.listdir(self._resDir)
		else:
			self._resDir = ""
			self._resFileNames = []

	def close(self) -> None:
		if self._file:
			try:
				self._file.close()
			except:
				log.exception("error while closing csv file")
		self.clear()

	def __len__(self) -> int:
		if self._wordCount is None:
			log.debug("Try not to use len(reader) as it takes extra time")
			self._wordCount = fileCountLines(self._filename) - \
				self._leadingLinesCount
		return self._wordCount + len(self._resFileNames)

	def __iter__(self) -> Iterator[BaseEntry]:
		if not self._csvReader:
			log.error("%s is not open, can not iterate", self)
			raise StopIteration

		wordCount = 0
		for row in self._csvReader:
			wordCount += 1
			if not row:
				yield None  # update progressbar
				continue
			try:
				word = row[0]
				defi = row[1]
			except IndexError:
				log.error("invalid row: %r", row)
				yield None  # update progressbar
				continue
			try:
				alts = row[2].split(",")
			except IndexError:
				pass
			else:
				word = [word] + alts
			yield self._glos.newEntry(word, defi)
		self._wordCount = wordCount

		resDir = self._resDir
		for fname in self._resFileNames:
			with open(join(resDir, fname), "rb") as fromFile:
				yield self._glos.newDataEntry(
					fname,
					fromFile.read(),
				)


def write(glos: GlossaryType, filename: str, encoding: str = "utf-8", resources: bool = True) -> None:
	resDir = filename + "_res"
	if not isdir(resDir):
		os.mkdir(resDir)
	with open(filename, "w", encoding=encoding) as csvfile:
		writer = csv.writer(
			csvfile,
			dialect="excel",
			quoting=csv.QUOTE_ALL,  # FIXME
		)
		for entry in glos:
			if entry.isData():
				if resources:
					entry.save(resDir)
				continue

			words = entry.getWords()
			if not words:
				continue
			word, alts = words[0], words[1:]
			defi = entry.getDefi()

			row = [
				word,
				defi,
			]
			if alts:
				row.append(",".join(alts))

			writer.writerow(row)

	if not os.listdir(resDir):
		os.rmdir(resDir)



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.


charsetByCode = {
	0x41: "cp1252",  # Default, 0x41
	0x42: "cp1252",  # Latin, 0x42
	0x43: "cp1250",  # Eastern European, 0x43
	0x44: "cp1251",  # Cyrillic, 0x44
	0x45: "cp932",  # Japanese, 0x45
	0x46: "cp950",  # Traditional Chinese, 0x46
	0x47: "cp936",  # Simplified Chinese, 0x47
	0x48: "cp1257",  # Baltic, 0x48
	0x49: "cp1253",  # Greek, 0x49
	0x4A: "cp949",  # Korean, 0x4A
	0x4B: "cp1254",  # Turkish, 0x4B
	0x4C: "cp1255",  # Hebrew, 0x4C
	0x4D: "cp1256",  # Arabic, 0x4D
	0x4E: "cp874",  # Thai, 0x4E
}



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = 'Omnidic'
description = 'Omnidic'
extensions = ['.omni', '.omnidic']
optionsProp = {
	"dicIndex": IntOption(),
}
depends = {}


def write(glos: GlossaryType, filename: str, dicIndex=16):
	if not isinstance(dicIndex, int):
		raise TypeError(
			'invalid dicIndex=%r, must be integer' % dicIndex
		)
	with indir(filename, create=True):
		indexFp = open(str(dicIndex), 'w')

		for bucketIndex, bucket in enumerate(glos.iterEntryBuckets(100)):
			if bucketIndex == 0:
				bucketFilename = '%s99' % dicIndex
			else:
				bucketFilename = '%s%s' % (
					dicIndex,
					bucketIndex * 100 + len(bucket) - 1,
				)

			indexFp.write('%s#%s#%s\n' % (
				bucket[0].getWord(),
				bucket[-1].getWord(),
				bucketFilename,
			))

			bucketFileObj = open(bucketFilename, 'w')
			for entry in bucket:
				word = entry.getWord()
				defi = entry.getDefi()
				defi = defi.replace('\n', '  ')  # FIXME
				bucketFileObj.write('%s#%s\n' % (word, defi))
			bucketFileObj.close()

		indexFp.close()



# -*- coding: utf-8 -*-
# dsl/__init__.py
# Read ABBYY Lingvo DSL dictionary format
#
# Copyright (C) 2013 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
# Copyright (C) 2013-2016 Saeed Rasooli <saeed.gnu@gmail.com>
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import re
import html.entities
from xml.sax.saxutils import escape, quoteattr

from formats_common import *
from . import flawless_dsl

enable = True
format = "ABBYYLingvoDSL"
description = "ABBYY Lingvo DSL (dsl)"
extensions = [".dsl"]
optionsProp = {
	"encoding": EncodingOption(),
	"audio": BoolOption(),
	"onlyFixMarkUp": BoolOption(),
}
depends = {}

__all__ = ["read"]

# {{{
# modified to work around codepoints that are not supported by `unichr`.
# http://effbot.org/zone/re-sub.htm#unescape-html
# January 15, 2003 | Fredrik Lundh


# Removes HTML or XML character references and entities from a text string.
#
# @param text The HTML (or XML) source text.
# @return The plain text, as a Unicode string, if necessary.

def unescape(text):
	def fixup(m):
		text = m.group(0)
		if text[:2] == "&#":
			# character reference
			try:
				if text[:3] == "&#x":
					i = int(text[3:-1], 16)
				else:
					i = int(text[2:-1])
			except ValueError:
				pass
			else:
				try:
					return chr(i)
				except ValueError:
					return ("\\U%08x" % i)\
						.decode("unicode-escape").encode("utf-8")
		else:
			# named entity
			try:
				text = chr(html.entities.name2codepoint[text[1:-1]])
			except KeyError:
				pass
		return text  # leave as is
	return re.sub("&#?\w+;", fixup, text)
# }}}


def make_a_href(s):
	return "<a href=%s>%s</a>" % (quoteattr(s), escape(s))


def ref_sub(x):
	return make_a_href(unescape(x.groups()[0]))


# order matters, a lot.
shortcuts = [
	# canonical: m > * > ex > i > c
	(
		r'[i][c](.*?)[/c][/i]',
		r'<i style="color:green">\g<1></i>'
	),
	(
		r'[m(\d)][ex](.*?)[/ex][/m]',
		r'<div class="ex" '
		r'style="margin-left:\g<1>em;color:steelblue">\g<2></div>'
	),
	(
		r'[m(\d)][*][ex](.*?)[/ex][/*][/m]',
		r'<div class="sec ex" '
		r'style="margin-left:\g<1>em;color:steelblue">\g<2></div>'
	),
	(
		r'[*][ex](.*?)[/ex][/*]',
		r'<span class="sec ex" style="color:steelblue">\g<1></span>'
	),
	(
		r'[m1](?:-{2,})[/m]',
		'<hr/>'
	),
	(
		r'[m(\d)](?:-{2,})[/m]',
		r'<hr style="margin-left:\g<1>em"/>'
	),
]

shortcuts = [
	(
		re.compile(repl.replace('[', r'\[').replace('*]', r'\*]')),
		sub
	) for (repl, sub) in shortcuts
]


# precompiled regexs
re_brackets_blocks = re.compile(r'\{\{[^}]*\}\}')
re_lang_open = re.compile(r'(?<!\\)\[lang[^\]]*\]')
re_m_open = re.compile(r'(?<!\\)\[m\d\]')
re_c_open_color = re.compile(r'\[c (\w+)\]')
re_sound = re.compile(r'\[s\]([^\[]*?)(wav|mp3)\s*\[/s\]')
re_img = re.compile(r'\[s\]([^\[]*?)(jpg|jpeg|gif|tif|tiff)\s*\[/s\]')
re_m = re.compile(r'\[m(\d)\](.*?)\[/m\]')
wrapped_in_quotes_re = re.compile(r'^(\'|")(.*)(\1)$')

# single instance of parser
# it's safe as long as this script's not going multithread.
_parse = flawless_dsl.FlawlessDSLParser().parse


def apply_shortcuts(line):
	for repl, sub in shortcuts:
		line = re.sub(repl, sub, line)
	return line


def _clean_tags(line, audio):
	r"""
	WARNING! shortcuts may apply:
		[m2][*][ex]{}[/ex][/*][/m]
		=>
		<div class="sec ex" style="margin-left:2em;color:steelblue">{}</div>
	[m{}] => <div style="margin-left:{}em">
	[*]   => <span class="sec">
	[ex]  => <span class="ex" style="color:steelblue">
	[c]   => <span style="color:green">
	[p]   => <i class="p" style="color:green">

	[']   => <u>
	[b]   => <b>
	[i]   => <i>
	[u]   => <u>
	[sup] => <sup>
	[sub] => <sub>

	[ref]   \
	[url]    } => <a href={}>{}</a>
	<<...>> /

	[s] =>  <object type="audio/x-wav" data="{}" width="40" height="40">
				<param name="autoplay" value="false" />
			</object>
	[s] =>  <img align="top" src="{}" alt="{}" />

	[t] => <!-- T --><span style="font-family:'Helvetica'">

	{{...}}   \
	[trn]      |
	[!trn]     |
	[trs]      } => remove
	[!trs]     |
	[lang ...] |
	[com]     /
	"""
	# remove {{...}} blocks
	line = re_brackets_blocks.sub("", line)
	# remove trn tags
	# re_trn = re.compile("\[\/?!?tr[ns]\]")
	line = line \
		.replace("[trn]", "") \
		.replace("[/trn]", "") \
		.replace("[trs]", "") \
		.replace("[/trs]", "") \
		.replace("[!trn]", "") \
		.replace("[/!trn]", "") \
		.replace("[!trs]", "") \
		.replace("[/!trs]", "")

	# remove lang tags
	line = re_lang_open.sub("", line).replace("[/lang]", "")
	# remove com tags
	line = line.replace("[com]", "").replace("[/com]", "")
	# remove t tags
	line = line.replace(
		"[t]",
		"<!-- T --><span style=\"font-family:'Helvetica'\">"
	)
	line = line.replace("[/t]", "</span><!-- T -->")

	line = _parse(line)

	line = re.sub(r"\\$", "<br/>", line)

	# paragraph, part one: before shortcuts.
	line = line.replace("[m]", "[m1]")
	# if line somewhere contains "[m_]" tag like
	# "[b]I[/b][m1] [c][i]conj.[/i][/c][/m][m1]1) ...[/m]"
	# then leave it alone.  only wrap in "[m1]" when no "m" tag found at all.
	if not re_m_open.search(line):
		line = "[m1]%s[/m]" % line

	line = apply_shortcuts(line)

	# paragraph, part two: if any not shourcuted [m] left?
	line = re_m.sub(r'<div style="margin-left:\g<1>em">\g<2></div>', line)

	# text formats

	line = line.replace("[']", '<u>').replace("[/']", '</u>')
	line = line.replace('[b]', '<b>').replace('[/b]', '</b>')
	line = line.replace('[i]', '<i>').replace('[/i]', '</i>')
	line = line.replace('[u]', '<u>').replace('[/u]', '</u>')
	line = line.replace('[sup]', '<sup>').replace('[/sup]', '</sup>')
	line = line.replace('[sub]', '<sub>').replace('[/sub]', '</sub>')

	# color
	line = line.replace('[c]', r'<span style="color:green">')
	line = re_c_open_color.sub(r'<span style="color:\g<1>">', line)
	line = line.replace('[/c]', r'</span>')

	# example zone
	line = line.replace('[ex]', r'<span class="ex" style="color:steelblue">')
	line = line.replace('[/ex]', r'</span>')

	# secondary zone
	line = line.replace('[*]', '<span class="sec">')\
		.replace('[/*]', '</span>')

	# abbrev. label
	line = line.replace('[p]', '<i class="p" style="color:green">')
	line = line.replace('[/p]', '</i>')

	# cross reference
	line = line.replace('[ref]', '<<').replace('[/ref]', '>>')
	line = line.replace('[url]', '<<').replace('[/url]', '>>')
	line = re.sub('<<(.*?)>>', ref_sub, line)

	# sound file
	if audio:
		sound_tag = r'<object type="audio/x-wav" data="\g<1>\g<2>" ' \
			'width="40" height="40">' \
			'<param name="autoplay" value="false" />' \
			'</object>'
	else:
		sound_tag = ""
	line = re_sound.sub(sound_tag, line)

	# image file
	line = re_img.sub(
		r'<img align="top" src="\g<1>\g<2>" alt="\g<1>\g<2>" />',
		line,
	)

	# \[...\]
	line = line.replace(r'\[', '[').replace(r'\]', ']')
	return line


def unwrap_quotes(s):
	return wrapped_in_quotes_re.sub(r'\2', s)


def read(
	glos: GlossaryType,
	filename: str,
	encoding: str = "",
	audio: bool = False,
	onlyFixMarkUp: bool = False,
):
	if onlyFixMarkUp:
		def clean_tags(line, audio):
			return _parse(line)
	else:
		clean_tags = _clean_tags

	def setInfo(key, value):
		glos.setInfo(key, unwrap_quotes(value))

	current_key = ""
	current_key_alters = []
	current_text = []
	line_type = "header"
	unfinished_line = ""

	if not encoding:
		for testEncoding in ("utf-8", "utf-16"):
			with open(filename, "r", encoding=testEncoding) as fp:
				try:
					for i in range(10):
						fp.readline()
				except UnicodeDecodeError:
					log.info("Encoding of DSL file is not %s" % testEncoding)
					continue
				else:
					log.info("Encoding of DSL file detected: %s" % testEncoding)
					encoding = testEncoding
					break
		if not encoding:
			raise ValueError("Could not detect encoding of DSL file, specify it by: --read-options encoding=ENCODING")

	fp = open(filename, "r", encoding=encoding)

	for line in fp:
		line = line.rstrip()
		if not line:
			continue
		# header
		if line.startswith("#"):
			if line.startswith("#NAME"):
				setInfo("title", line[6:])
			elif line.startswith("#INDEX_LANGUAGE"):
				setInfo("sourceLang", line[16:])
			elif line.startswith("#CONTENTS_LANGUAGE"):
				setInfo("targetLang", line[19:])
			line_type = "header"
		# texts
		elif line.startswith(" ") or line.startswith("\t"):
			line_type = "text"
			line = unfinished_line + line.lstrip()

			# some ill formated source may have tags spanned into
			# multiple lines
			# try to match opening and closing tags
			tags_open = re.findall(r'(?<!\\)\[(c |[cuib]\])', line)
			tags_close = re.findall(r'\[/[cuib]\]', line)
			if len(tags_open) != len(tags_close):
				unfinished_line = line
				continue

			unfinished_line = ""

			# convert DSL tags to HTML tags
			line = clean_tags(line, audio)
			current_text.append(line)
		# title word(s)
		else:
			# alternative titles
			if line_type == "title":
				current_key_alters.append(line)
			# previous line type is text -> start new title
			else:
				# append previous entry
				if line_type == "text":
					if unfinished_line:
						# line may be skipped if ill formated
						current_text.append(clean_tags(unfinished_line, audio))
					glos.addEntry(
						[current_key] + current_key_alters,
						"\n".join(current_text),
					)
				# start new entry
				current_key = line
				current_key_alters = []
				current_text = []
				unfinished_line = ""
			line_type = "title"
	fp.close()

	# last entry
	if line_type == "text":
		glos.addEntry(
			[current_key] + current_key_alters,
			"\n".join(current_text),
		)



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import re
from pyglossary.plugins.formats_common import log
from pyglossary.xml_utils import xml_escape


u_pat_html_entry = re.compile("(?:&#x|&#|&)(\\w+);?", re.I)
u_pat_html_entry_key = re.compile("(?:&#x|&#|&)(\\w+);", re.I)
b_pat_ascii_char_ref = re.compile(b"(&#\\w+;)", re.I)

unkownHtmlEntries = set()


def replaceHtmlEntryNoEscapeCB(u_match):
	"""
	u_match: instance of _sre.SRE_Match
	Replace character entity with the corresponding character

	Return the original string if conversion fails.
	Use this as a replace function of re.sub.
	"""
	import html.entities
	from pyglossary.html_utils import name2codepoint

	u_text = u_match.group(0)
	u_name = u_match.group(1)
	if log.isDebug():
		assert isinstance(u_text, str) and isinstance(u_name, str)

	u_res = None
	if u_text[:2] == "&#":
		# character reference
		try:
			if u_text[:3].lower() == "&#x":
				code = int(u_name, 16)
			else:
				code = int(u_name)
			if code <= 0:
				raise ValueError()
			u_res = chr(code)
		except (ValueError, OverflowError):
			u_res = chr(0xFFFD)  # replacement character
	elif u_text[0] == "&":
		# named entity
		try:
			u_res = chr(html.entities.name2codepoint[u_name])
		except KeyError:
			try:
				u_res = chr(name2codepoint[u_name.lower()])
			except KeyError:
				"""
				Babylon dictionaries contain a lot of non-standard entity,
				references for example, csdot, fllig, nsm, cancer, thlig,
				tsdot, upslur...
				This not just a typo. These entries repeat over and over again.
				Perhaps they had meaning in the source dictionary that was
				converted to Babylon, but now the meaning is lost. Babylon
				does render them as is, that is, for example, &csdot; despite
				other references like &amp; are replaced with corresponding
				characters.
				"""
				unkownHtmlEntries.add(u_text)
				u_res = u_text
	else:
		raise ArgumentError()
	return u_res


def replaceHtmlEntryCB(u_match):
	"""
	u_match: instance of _sre.SRE_Match
	Same as replaceHtmlEntryNoEscapeCB, but escapes result string

	Only <, >, & characters are escaped.
	"""
	u_res = replaceHtmlEntryNoEscapeCB(u_match)
	if u_match.group(0) == u_res:  # conversion failed
		return u_res
	else:
		return xml_escape(u_res)


def replaceDingbat(u_match):
	"""
	u_match: instance of _sre.SRE_Match
	replace chars \\u008c-\\u0095 with \\u2776-\\u277f
	"""
	ch = u_match.group(0)
	code = ch + 0x2776 - 0x8c
	return chr(code)


def escapeNewlinesCallback(u_match):
	"""
	u_match: instance of _sre.SRE_Match
	"""
	ch = u_match.group(0)
	if ch == "\n":
		return "\\n"
	if ch == "\r":
		return "\\r"
	if ch == "\\":
		return "\\\\"
	return ch


def replaceHtmlEntries(u_text):
	# &ldash;
	# &#0147;
	# &#x010b;
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		u_pat_html_entry,
		replaceHtmlEntryCB,
		u_text,
	)


def replaceHtmlEntriesInKeys(u_text):
	# &ldash;
	# &#0147;
	# &#x010b;
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		u_pat_html_entry_key,
		replaceHtmlEntryNoEscapeCB,
		u_text,
	)


def escapeNewlines(u_text):
	r"""
	convert text to c-escaped string:
	\ -> \\
	new line -> \n or \r
	"""
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		"[\\r\\n\\\\]",
		escapeNewlinesCallback,
		u_text,
	)


def stripHtmlTags(u_text):
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		"(?:<[/a-zA-Z].*?(?:>|$))+",
		" ",
		u_text,
	)


def removeControlChars(u_text):
	# \x09 - tab
	# \x0a - line feed
	# \x0b - vertical tab
	# \x0d - carriage return
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		"[\x00-\x08\x0c\x0e-\x1f]",
		"",
		u_text,
	)


def removeNewlines(u_text):
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		"[\r\n]+",
		" ",
		u_text,
	)


def normalizeNewlines(u_text):
	"""
	convert new lines to unix style and remove consecutive new lines
	"""
	if log.isDebug():
		assert isinstance(u_text, str)
	return re.sub(
		"[\r\n]+",
		"\n",
		u_text,
	)


def replaceAsciiCharRefs(b_text, encoding):
	# &#0147;
	# &#x010b;
	if log.isDebug():
		assert isinstance(b_text, bytes)
	b_parts = re.split(b_pat_ascii_char_ref, b_text)
	for i_part, b_part in enumerate(b_parts):
		if i_part % 2 != 1:
			continue
		# reference
		try:
			if b_part[:3].lower() == "&#x":
				code = int(b_part[3:-1], 16)
			else:
				code = int(b_part[2:-1])
			if code <= 0:
				raise ValueError()
		except (ValueError, OverflowError):
			code = -1
		if code < 128 or code > 255:
			continue
		# no need to escape "<", ">", "&"
		b_parts[i_part] = bytes([code])
	return b"".join(b_parts)


def fixImgLinks(u_text):
	"""
	Fix img tag links

	src attribute value of image tag is often enclosed in \x1e - \x1f
	characters.
	For example:
		<IMG border='0' src='\x1e6B6C56EC.png\x1f' width='9' height='8'>.
	Naturally the control characters are not part of the image source name.
	They may be used to quickly find all names of resources.
	This function strips all such characters.
	Control characters \x1e and \x1f are useless in html text, so we may
	safely remove all of them, irrespective of context.
	"""
	if log.isDebug():
		assert isinstance(u_text, str)
	return u_text.replace("\x1e", "").replace("\x1f", "")


def stripDollarIndexes(b_word):
	if log.isDebug():
		assert isinstance(b_word, bytes)
	i = 0
	b_word_main = b""
	strip_count = 0  # number of sequences found
	# strip $<index>$ sequences
	while True:
		d0 = b_word.find(b"$", i)
		if d0 == -1:
			b_word_main += b_word[i:]
			break
		d1 = b_word.find(b"$", d0 + 1)
		if d1 == -1:
			# log.debug(
			#	"stripDollarIndexes(%s):\npaired $ is not found" % b_word
			# )
			b_word_main += b_word[i:]
			break
		if d1 == d0 + 1:
			"""
			You may find keys (or alternative keys) like these:
			sur l'arbre$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
			obscurantiste$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
			They all end on a sequence of b'$', key length including dollars
			is always 60 chars.
			You may find keys like these:
			extremidade-$$$-$$$-linha
			.FIRM$$$$$$$$$$$$$
			etc

			summary: we must remove any sequence of dollar signs longer
			than 1 chars
			"""
			# log.debug("stripDollarIndexes(%s):\nfound $$", b_word)
			b_word_main += b_word[i:d0]
			i = d1 + 1
			while i < len(b_word) and b_word[i] == ord(b"$"):
				i += 1
			if i >= len(b_word):
				break
			continue
		if b_word[d0 + 1:d1].strip(b"0123456789"):
			# if has at least one non-digit char
			# log.debug("stripDollarIndexes(%s):\nnon-digit between $$', b_word)
			b_word_main += b_word[i:d1]
			i = d1
			continue
		if d1 + 1 < len(b_word) and b_word[d1 + 1] != 0x20:
			"""
			Examples:
		make do$4$/make /do
		potere$1$<BR><BR>
		See also <a href='file://ITAL-ENG POTERE 1249-1250.pdf'>notes...</A>
		volere$1$<BR><BR>
		See also <a href='file://ITAL-ENG VOLERE 1469-1470.pdf'>notes...</A>
		Ihre$1$Ihres
			"""
			log.debug(
				"stripDollarIndexes(%s):\n" % b_word +
				"second $ is followed by non-space"
			)
			pass
		b_word_main += b_word[i:d0]
		i = d1 + 1
		strip_count += 1

	return b_word_main, strip_count



# -*- coding: utf-8 -*-
# ui_gtk.py
#
# Copyright © 2008-2019 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Thanks to Pier Carteri <m3tr0@dei.unipd.it> for program Py_Shell.py
# Thanks to Milad Rastian for program pySQLiteGUI
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import shutil
import sys
import os
from os.path import join, isfile, isabs, splitext
import logging
import traceback

from typing import (
	List,
	Dict,
	Any,
)

from pyglossary.text_utils import urlToPath
from pyglossary.os_utils import click_website
from pyglossary.glossary import *
from .base import *
from pyglossary import core
from .dependency import checkFormatDepends

import gi
gi.require_version("Gtk", "3.0")

from .gtk3_utils import *
from .gtk3_utils.utils import *
from .gtk3_utils.dialog import MyDialog
from .gtk3_utils.resize_button import ResizeButton

# from gi.repository import GdkPixbuf

log = logging.getLogger("root")

gtk.Window.set_default_icon_from_file(logo)

_ = str  # later replace with translator function

pixDir = join(dataDir, "res")  # FIXME


def getCopressedFileExt(fpath):
	fname, ext = splitext(fpath.lower())
	if ext in (".gz", ".bz2", ".zip"):
		fname, ext = splitext(fname)
	return ext


def buffer_get_text(b):
	return b.get_text(
		b.get_start_iter(),
		b.get_end_iter(),
		True,
	)


class FormatOptionsDialog(gtk.Dialog):
	def __init__(self, formatName: str, options: List[str], optionsValues: Dict[str, Any], parent=None):
		gtk.Dialog.__init__(self, parent=parent)
		optionsProp = Glossary.formatsOptionsProp[formatName]
		self.optionsProp = optionsProp
		##
		self.connect("response", lambda w, e: self.hide())
		dialog_add_button(
			self,
			"gtk-ok",
			"_OK",
			gtk.ResponseType.OK,
		)
		dialog_add_button(
			self,
			"gtk-cancel",
			"_Cancel",
			gtk.ResponseType.CANCEL,
		)
		###
		treev = gtk.TreeView()
		trees = gtk.ListStore(
			bool, # enable
			str, # name
			str, # comment
			str, # value
		)
		treev.set_headers_clickable(True)
		treev.set_model(trees)
		treev.connect("row-activated", self.rowActivated)
		treev.connect("button-press-event", self.treeviewButtonPress)
		###
		self.treev = treev
		#############
		cell = gtk.CellRendererToggle()
		#cell.set_property("activatable", True)
		cell.connect("toggled", self.enableToggled)
		col = gtk.TreeViewColumn(title="Enable", cell_renderer=cell)
		col.add_attribute(cell, "active", 0)
		#cell.set_active(False)
		col.set_property("expand", False)
		col.set_resizable(True)
		treev.append_column(col)
		###
		col = gtk.TreeViewColumn(title="Name", cell_renderer=gtk.CellRendererText(), text=1)
		col.set_property("expand", False)
		col.set_resizable(True)
		treev.append_column(col)
		###
		cell = gtk.CellRendererText(editable=True)
		self.valueCell = cell
		self.valueCol = 3
		cell.connect("edited", self.valueEdited)
		col = gtk.TreeViewColumn(title="Value", cell_renderer=cell, text=self.valueCol)
		col.set_property("expand", True)
		col.set_resizable(True)
		col.set_min_width(200)
		treev.append_column(col)
		###
		col = gtk.TreeViewColumn(title="Comment", cell_renderer=gtk.CellRendererText(), text=2)
		col.set_property("expand", False)
		col.set_resizable(False)
		treev.append_column(col)
		#############
		for name in options:
			prop = optionsProp[name]
			comment = prop.typ
			if prop.comment:
				comment += ", " + prop.comment
			if prop.typ == "str" and not prop.values:
				comment += ", double-click to edit"
			trees.append([
				name in optionsValues, # enable
				name, # name
				comment, # comment
				str(optionsValues.get(name, "")), # value
			])
		############
		pack(self.vbox, treev, 1, 1)
		self.vbox.show_all()

	def enableToggled(self, cell, path):
		# enable is column 0
		model = self.treev.get_model()
		active = not cell.get_active()
		itr = model.get_iter(path)
		model.set_value(itr, 0, active)

	def valueEdited(self, cell, path, rawValue):
		# value is column 3
		model = self.treev.get_model()
		itr = model.get_iter(path)
		optName = model.get_value(itr, 1)
		prop = self.optionsProp[optName]
		if not prop.customValue:
			return
		enable = True
		if rawValue == "" and prop.typ != "str":
			enable = False
		elif not prop.validateRaw(rawValue):
			log.error(f"invalid {prop.typ} value: {optName} = {rawValue!r}")
			return
		model.set_value(itr, self.valueCol, rawValue)
		model.set_value(itr, 0, enable)

	def rowActivated(self, treev, path, col):
		# forceMenu=True because we can not enter edit mode
		# if double-clicked on a cell other than Value
		return self.valueCellClicked(path, forceMenu=True)

	def treeviewButtonPress(self, treev, gevent):
		if gevent.button != 1:
			return False
		pos_t = treev.get_path_at_pos(int(gevent.x), int(gevent.y))
		if not pos_t:
			return False
		# pos_t == path, col, xRel, yRel
		path = pos_t[0]
		col = pos_t[1]
		# cell = col.get_cells()[0]
		if col.get_title() == "Value":
			return self.valueCellClicked(path)
		return False

	def valueItemActivate(self, item: gtk.MenuItem, itr: gtk.TreeIter):
		# value is column 3
		value = item.get_label()
		model = self.treev.get_model()
		model.set_value(itr, self.valueCol, value)
		model.set_value(itr, 0, True) # enable it

	def valueCustomOpenDialog(self, itr: gtk.TreeIter, optName: str):
		model = self.treev.get_model()
		prop = self.optionsProp[optName]
		currentValue = model.get_value(itr, self.valueCol)
		optDesc = optName
		if prop.comment:
			optDesc += " (%s)" % prop.comment
		label = gtk.Label(label="Value for %s" % (optDesc,))
		dialog = gtk.Dialog(parent=self, title="Option Value")
		dialog.connect("response", lambda w, e: dialog.hide())
		dialog_add_button(
			dialog,
			"gtk-ok",
			"_OK",
			gtk.ResponseType.OK,
		)
		dialog_add_button(
			dialog,
			"gtk-cancel",
			"_Cancel",
			gtk.ResponseType.CANCEL,
		)
		pack(dialog.vbox, label, 0, 0)
		entry = gtk.Entry()
		entry.set_text(currentValue)
		entry.connect("activate", lambda w: dialog.response(gtk.ResponseType.OK))
		pack(dialog.vbox, entry, 0, 0)
		dialog.vbox.show_all()
		if dialog.run() != gtk.ResponseType.OK:
			return
		value = entry.get_text()
		model.set_value(itr, self.valueCol, value)
		model.set_value(itr, 0, True) # enable it

	def valueItemCustomActivate(self, item: gtk.MenuItem, itr: gtk.TreeIter):
		model = self.treev.get_model()
		optName = model.get_value(itr, 1)
		self.valueCustomOpenDialog(itr, optName)

	def valueCellClicked(self, path, forceMenu=False) -> bool:
		"""
		returns True if event is handled, False if not handled (need to enter edit mode)
		"""
		model = self.treev.get_model()
		itr = model.get_iter(path)
		optName = model.get_value(itr, 1)
		prop = self.optionsProp[optName]
		if prop.typ == "bool":
			rawValue = model.get_value(itr, self.valueCol)
			if rawValue == "":
				value = False
			else:
				value, isValid = prop.evaluate(rawValue)
				if not isValid:
					log.error("invalid %s = %r" % (optName, rawValue))
					value = False
			model.set_value(itr, self.valueCol, str(not value))
			model.set_value(itr, 0, True) # enable it
			return True
		propValues = prop.values
		if not propValues:
			if forceMenu:
				propValues = []
			else:
				return False
		menu = gtk.Menu()
		if prop.customValue:
			item = gtk.MenuItem("[Custom Value]")
			item.connect("activate", self.valueItemCustomActivate, itr)
			item.show()
			menu.append(item)
		groupedValues = None
		if len(propValues) > 10:
			groupedValues = prop.groupValues()
		if groupedValues:
			for groupName, values in groupedValues.items():
				item = gtk.MenuItem()
				item.set_label(groupName)
				if values is None:
					item.connect("activate", self.valueItemActivate, itr)
				else:
					subMenu = gtk.Menu()
					for subValue in values:
						subItem = gtk.MenuItem(label=str(subValue))
						subItem.connect("activate", self.valueItemActivate, itr)
						subItem.show()
						subMenu.append(subItem)
					item.set_submenu(subMenu)
				item.show()
				menu.append(item)
		else:
			for value in propValues:
				item = gtk.MenuItem(value)
				item.connect("activate", self.valueItemActivate, itr)
				item.show()
				menu.append(item)
		etime = gtk.get_current_event_time()
		menu.popup(None, None, None, None, 3, etime)
		return True

	def getOptionsValues(self):
		model = self.treev.get_model()
		optionsValues = {}
		for row in model:
			if not row[0]: # not enable
				continue
			optName = row[1]
			rawValue = row[3]
			prop = self.optionsProp[optName]
			value, isValid = prop.evaluate(rawValue)
			if not isValid:
				log.error("invalid option value %s = %s" % (optName, rawValue))
				continue
			optionsValues[optName] = value
		return optionsValues



class FormatComboBox(gtk.ComboBox):
	def __init__(self, parent=None):
		gtk.ComboBox.__init__(self)
		self._parent = parent
		self.model = gtk.ListStore(
			str,  # format name, hidden
			# GdkPixbuf.Pixbuf,# icon
			str,  # format description, shown
		)
		self.set_model(self.model)

		cell = gtk.CellRendererText()
		cell.set_visible(False)
		pack(self, cell)
		self.add_attribute(cell, "text", 0)

		# cell = gtk.CellRendererPixbuf()
		# pack(self, cell, False)
		# self.add_attribute(cell, "pixbuf", 1)

		cell = gtk.CellRendererText()
		pack(self, cell, True)
		self.add_attribute(cell, "text", 1)

		self.optionsValues = {}

		self.optionsButton = gtk.Button(label="Options")
		self.optionsButton.connect("clicked", self.optionsButtonClicked)

		self.dependsButton = gtk.Button(label="Install dependencies")
		self.dependsButton.pkgNames = []
		self.dependsButton.connect("clicked", self.dependsButtonClicked)

		self.connect("changed", self.onChanged)

	def setOptionsValues(self, optionsValues: Dict[str, Any]):
		self.optionsValues = optionsValues

	def kind(self):
		"returns 'r' or 'w'"
		raise NotImplementedError

	def getActiveOptions(self):
		raise NotImplementedError

	def addFormat(self, _format):
		self.get_model().append((
			_format,
			# icon,
			Glossary.formatsDesc[_format],
		))

	def getActive(self):
		index = gtk.ComboBox.get_active(self)
		if index is None or index < 0:
			return ""
		return self.get_model()[index][0]

	def setActive(self, _format):
		for i, row in enumerate(self.get_model()):
			if row[0] == _format:
				gtk.ComboBox.set_active(self, i)
				return

	def optionsButtonClicked(self, button):
		formatName = self.getActive()
		options = self.getActiveOptions()
		dialog = FormatOptionsDialog(
			formatName,
			options,
			self.optionsValues,
			parent=self._parent,
		)
		dialog.set_title("Options for " + formatName)
		if dialog.run() != gtk.ResponseType.OK:
			dialog.destroy()
			return
		self.optionsValues = dialog.getOptionsValues()
		dialog.destroy()

	def dependsButtonClicked(self, button):
		formatName = self.getActive()
		pkgNames = button.pkgNames
		if not pkgNames:
			print("All dependencies are stattisfied for " + formatName)
			return
		pkgNamesStr = " ".join(pkgNames)
		msg = "\n".join([
			"Run the following command:",
			"sudo pip3 install " + pkgNamesStr,
		])
		showInfo(
			msg,
			title="Dependencies for " + formatName,
			selectable=True,
			parent=self._parent,
		)
		self.onChanged(self)

	def onChanged(self, combo):
		name = self.getActive()
		self.optionsValues.clear()

		options = self.getActiveOptions()
		self.optionsButton.set_visible(bool(options))

		uninstalled = checkFormatDepends(name)
		self.dependsButton.pkgNames = uninstalled
		self.dependsButton.set_visible(bool(uninstalled))


class InputFormatComboBox(FormatComboBox):
	def __init__(self, **kwargs):
		FormatComboBox.__init__(self, **kwargs)
		for _format in Glossary.readFormats:
			self.addFormat(_format)

	def kind(self):
		"returns 'r' or 'w'"
		return "r"

	def getActiveOptions(self):
		return Glossary.formatsReadOptions[self.getActive()]


class OutputFormatComboBox(FormatComboBox):
	def __init__(self, **kwargs):
		FormatComboBox.__init__(self, **kwargs)
		for _format in Glossary.writeFormats:
			self.addFormat(_format)

	def kind(self):
		"returns 'r' or 'w'"
		return "w"

	def getActiveOptions(self):
		return Glossary.formatsWriteOptions[self.getActive()]


class GtkTextviewLogHandler(logging.Handler):
	def __init__(self, treeview_dict):
		logging.Handler.__init__(self)

		self.buffers = {}
		for levelname in (
			"CRITICAL",
			"ERROR",
			"WARNING",
			"INFO",
			"DEBUG",
		):
			textview = treeview_dict[levelname]

			buff = textview.get_buffer()
			tag = gtk.TextTag.new(levelname)
			buff.get_tag_table().add(tag)

			self.buffers[levelname] = buff

	def getTag(self, levelname):
		return self.buffers[levelname].get_tag_table().lookup(levelname)

	def setColor(self, levelname: str, rgba: gdk.RGBA) -> None:
		self.getTag(levelname).set_property("foreground-rgba", rgba)
		# foreground-gdk is deprecated since Gtk 3.4

	def emit(self, record):
		msg = record.getMessage()
		# msg = msg.replace("\x00", "")

		if record.exc_info:
			_type, value, tback = record.exc_info
			tback_text = "".join(
				traceback.format_exception(_type, value, tback)
			)
			if msg:
				msg += "\n"
			msg += tback_text

		buff = self.buffers[record.levelname]

		buff.insert_with_tags_by_name(
			buff.get_end_iter(),
			msg + "\n",
			record.levelname,
		)


class GtkSingleTextviewLogHandler(GtkTextviewLogHandler):
	def __init__(self, textview):
		GtkTextviewLogHandler.__init__(self, {
			"CRITICAL": textview,
			"ERROR": textview,
			"WARNING": textview,
			"INFO": textview,
			"DEBUG": textview,
		})


class BrowseButton(gtk.Button):
	def __init__(
		self,
		setFilePathFunc,
		label="Browse",
		actionSave=False,
		title="Select File",
	):
		gtk.Button.__init__(self)

		self.set_label(label)
		self.set_image(gtk.Image.new_from_icon_name(
			"document-save" if actionSave else "document-open",
			gtk.IconSize.BUTTON,
		))

		self.actionSave = actionSave
		self.setFilePathFunc = setFilePathFunc
		self.title = title

		self.connect("clicked", self.onClick)

	def onClick(self, widget):
		fcd = gtk.FileChooserDialog(
			transient_for=self.get_toplevel(),
			action=gtk.FileChooserAction.SAVE if self.actionSave
			else gtk.FileChooserAction.OPEN,
			title=self.title,
		)
		fcd.add_button(gtk.STOCK_CANCEL, gtk.ResponseType.CANCEL)
		fcd.add_button(gtk.STOCK_OK, gtk.ResponseType.OK)
		fcd.connect("response", lambda w, e: fcd.hide())
		fcd.connect(
			"file-activated",
			lambda w: fcd.response(gtk.ResponseType.OK)
		)
		if fcd.run() == gtk.ResponseType.OK:
			self.setFilePathFunc(fcd.get_filename())
		fcd.destroy()


class UI(gtk.Dialog, MyDialog, UIBase):
	def write(self, tag):  # FIXME
		pass

	def status(self, msg):
		# try:
		#	_id = self.statusMsgDict[msg]
		# except KeyError:
		#	_id = self.statusMsgDict[msg] = self.statusNewId
		#	self.statusNewId += 1
		_id = self.statusBar.get_context_id(msg)
		self.statusBar.push(_id, msg)

	def __init__(self, **options):
		gtk.Dialog.__init__(self)
		self.set_title("PyGlossary (Gtk3)")
		self.resize(800, 800)
		self.connect("delete-event", self.onDeleteEvent)
		self.prefPages = []
		# self.statusNewId = 0
		# self.statusMsgDict = {}## message -> id
		#####
		self.pref = {}
		self.pref_load(**options)
		#####
		self.assert_quit = False
		self.path = ""
		self.glos = Glossary(ui=self)
		# ____________________ Tab 1 - Convert ____________________ #
		sizeGroup = gtk.SizeGroup(mode=gtk.SizeGroupMode.HORIZONTAL)
		####
		vbox = gtk.VBox()
		vbox.label = _("Convert")
		vbox.icon = ""  # "*.png"
		self.prefPages.append(vbox)
		######
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertInputFormatCombo = InputFormatComboBox(parent=self)
		pack(hbox, self.convertInputFormatCombo)
		pack(hbox, gtk.Label(), 1, 1)
		pack(hbox, self.convertInputFormatCombo.dependsButton)
		pack(hbox, self.convertInputFormatCombo.optionsButton)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertInputEntry = gtk.Entry()
		pack(hbox, self.convertInputEntry, 1, 1)
		button = BrowseButton(
			self.convertInputEntry.set_text,
			label="Browse",
			actionSave=False,
			title="Select Input File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.convertInputEntry.connect(
			"changed",
			self.convertInputEntryChanged,
		)
		#####
		vbox.sep1 = gtk.Label(label="")
		vbox.sep1.show()
		pack(vbox, vbox.sep1)
		#####
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertOutputFormatCombo = OutputFormatComboBox(parent=self)
		pack(hbox, self.convertOutputFormatCombo)
		pack(hbox, gtk.Label(), 1, 1)
		pack(hbox, self.convertOutputFormatCombo.dependsButton)
		pack(hbox, self.convertOutputFormatCombo.optionsButton)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.convertOutputEntry = gtk.Entry()
		pack(hbox, self.convertOutputEntry, 1, 1)
		button = BrowseButton(
			self.convertOutputEntry.set_text,
			label="Browse",
			actionSave=True,
			title="Select Output File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.convertOutputEntry.connect(
			"changed",
			self.convertOutputEntryChanged,
		)
		#####
		hbox = gtk.HBox(spacing=10)
		label = gtk.Label(label="")
		pack(hbox, label, 1, 1, 10)
		self.convertButton = gtk.Button()
		self.convertButton.set_label("Convert")
		self.convertButton.connect("clicked", self.convertClicked)
		pack(hbox, self.convertButton, 1, 1, 10)
		pack(vbox, hbox, 0, 0, 15)
		# ____________________ Tab 2 - Reverse ____________________ #
		self.reverseStatus = ""
		####
		sizeGroup = gtk.SizeGroup(mode=gtk.SizeGroupMode.HORIZONTAL)
		####
		vbox = gtk.VBox()
		vbox.label = _("Reverse")
		vbox.icon = ""  # "*.png"
		# self.prefPages.append(vbox)
		######
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input Format")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseInputFormatCombo = InputFormatComboBox()
		pack(hbox, self.reverseInputFormatCombo)
		pack(vbox, hbox)
		###
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Input File")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseInputEntry = gtk.Entry()
		pack(hbox, self.reverseInputEntry, 1, 1)
		button = BrowseButton(
			self.reverseInputEntry.set_text,
			label="Browse",
			actionSave=False,
			title="Select Input File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.reverseInputEntry.connect(
			"changed",
			self.reverseInputEntryChanged,
		)
		#####
		vbox.sep1 = gtk.Label(label="")
		vbox.sep1.show()
		pack(vbox, vbox.sep1)
		#####
		hbox = gtk.HBox(spacing=3)
		hbox.label = gtk.Label(label=_("Output Tabfile")+":")
		pack(hbox, hbox.label)
		sizeGroup.add_widget(hbox.label)
		hbox.label.set_property("xalign", 0)
		self.reverseOutputEntry = gtk.Entry()
		pack(hbox, self.reverseOutputEntry, 1, 1)
		button = BrowseButton(
			self.reverseOutputEntry.set_text,
			label="Browse",
			actionSave=True,
			title="Select Output File",
		)
		pack(hbox, button)
		pack(vbox, hbox)
		##
		self.reverseOutputEntry.connect(
			"changed",
			self.reverseOutputEntryChanged,
		)
		#####
		hbox = gtk.HBox(spacing=3)
		label = gtk.Label(label="")
		pack(hbox, label, 1, 1, 5)
		###
		self.reverseStartButton = gtk.Button()
		self.reverseStartButton.set_label(_("Start"))
		self.reverseStartButton.connect("clicked", self.reverseStartClicked)
		pack(hbox, self.reverseStartButton, 1, 1, 2)
		###
		self.reversePauseButton = gtk.Button()
		self.reversePauseButton.set_label(_("Pause"))
		self.reversePauseButton.set_sensitive(False)
		self.reversePauseButton.connect("clicked", self.reversePauseClicked)
		pack(hbox, self.reversePauseButton, 1, 1, 2)
		###
		self.reverseResumeButton = gtk.Button()
		self.reverseResumeButton.set_label(_("Resume"))
		self.reverseResumeButton.set_sensitive(False)
		self.reverseResumeButton.connect("clicked", self.reverseResumeClicked)
		pack(hbox, self.reverseResumeButton, 1, 1, 2)
		###
		self.reverseStopButton = gtk.Button()
		self.reverseStopButton.set_label(_("Stop"))
		self.reverseStopButton.set_sensitive(False)
		self.reverseStopButton.connect("clicked", self.reverseStopClicked)
		pack(hbox, self.reverseStopButton, 1, 1, 2)
		###
		pack(vbox, hbox, 0, 0, 5)
		#####
		# ____________________________________________________________ #
		notebook = gtk.Notebook()
		self.notebook = notebook
		#########
		for vbox in self.prefPages:
			l = gtk.Label(label=vbox.label)
			l.set_use_underline(True)
			vb = gtk.VBox(spacing=3)
			if vbox.icon:
				vbox.image = imageFromFile(vbox.icon)
				pack(vb, vbox.image)
			pack(vb, l)
			vb.show_all()
			notebook.append_page(vbox, vb)
			try:
				notebook.set_tab_reorderable(vbox, True)
			except AttributeError:
				pass
		#######################
		# notebook.set_property("homogeneous", True)  # not in gtk3 FIXME
		# notebook.set_property("tab-border", 5)  # not in gtk3 FIXME
		# notebook.set_property("tab-hborder", 15)  # not in gtk3 FIXME
		pack(self.vbox, notebook, 0, 0)
		# for i in ui.prefPagesOrder:
		#	try:
		#		j = prefPagesOrder[i]
		#	except IndexError:
		#		continue
		#	notebook.reorder_child(self.prefPages[i], j)
		# ____________________________________________________________ #
		self.consoleTextview = textview = gtk.TextView()
		swin = gtk.ScrolledWindow()
		swin.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
		swin.set_border_width(0)
		swin.add(textview)
		pack(self.vbox, swin, 1, 1)
		###
		handler = GtkSingleTextviewLogHandler(textview)
		log.addHandler(handler)
		###
		textview.override_background_color(
			gtk.StateFlags.NORMAL,
			gdk.RGBA(0, 0, 0, 1),
		)
		###
		handler.setColor("CRITICAL", rgba_parse("red"))
		handler.setColor("ERROR", rgba_parse("red"))
		handler.setColor("WARNING", rgba_parse("yellow"))
		handler.setColor("INFO", rgba_parse("white"))
		handler.setColor("DEBUG", rgba_parse("white"))
		###
		textview.get_buffer().set_text("Output & Error Console:\n")
		textview.set_editable(False)
		# ____________________________________________________________ #
		self.progressTitle = ""
		self.progressBar = pbar = gtk.ProgressBar()
		pbar.set_fraction(0)
		# pbar.set_text(_("Progress Bar"))
		# pbar.get_style_context()
		# pbar.set_property("height-request", 20)
		pack(self.vbox, pbar, 0, 0)
		############
		hbox = gtk.HBox(spacing=5)
		clearButton = gtk.Button(
			use_stock=gtk.STOCK_CLEAR,
			always_show_image=True,
			label=_("Clear"),
		)
		clearButton.show_all()
		# image = gtk.Image()
		# image.set_from_stock(gtk.STOCK_CLEAR, gtk.IconSize.MENU)
		# clearButton.add(image)
		clearButton.set_border_width(0)
		clearButton.connect("clicked", self.consoleClearButtonClicked)
		set_tooltip(clearButton, "Clear Console")
		pack(hbox, clearButton, 0, 0)
		####
		# hbox.sepLabel1 = gtk.Label(label="")
		# pack(hbox, hbox.sepLabel1, 1, 1)
		######
		hbox.verbosityLabel = gtk.Label(label=_("Verbosity")+":")
		pack(hbox, hbox.verbosityLabel, 0, 0)
		##
		self.verbosityCombo = combo = gtk.ComboBoxText()
		for level, levelName in enumerate(log.levelNamesCap):
			combo.append_text("%s - %s" % (
				level,
				_(levelName)
			))
		combo.set_active(log.getVerbosity())
		combo.set_border_width(0)
		combo.connect("changed", self.verbosityComboChanged)
		pack(hbox, combo, 0, 0)
		####
		# hbox.sepLabel2 = gtk.Label(label="")
		# pack(hbox, hbox.sepLabel2, 1, 1)
		####
		self.statusBar = sbar = gtk.Statusbar()
		pack(hbox, self.statusBar, 1, 1)
		####
		hbox.resizeButton = ResizeButton(self)
		pack(hbox, hbox.resizeButton, 0, 0)
		######
		pack(self.vbox, hbox, 0, 0)
		# ____________________________________________________________ #
		self.vbox.show_all()
		self.convertInputFormatCombo.dependsButton.hide()
		self.convertOutputFormatCombo.dependsButton.hide()
		self.convertInputFormatCombo.optionsButton.hide()
		self.convertOutputFormatCombo.optionsButton.hide()
		########
		self.status("Select input file")

	def run(self, editPath=None, readOptions=None):
		if readOptions is not None:
			self.convertInputFormatCombo.setOptionsValues(readOptions)
		# if editPath:
		#	self.notebook.set_current_page(3)
		#	log.info("Opening file \"%s\" for edit. please wait...", editPath)
		#	while gtk.events_pending():
		#		gtk.main_iteration_do(False)
		#	self.dbe_open(editPath, **readOptions)
		gtk.Dialog.present(self)
		gtk.main()

	def onDeleteEvent(self, widget, event):
		self.destroy()
		# gtk.main_quit()
		# if callled while converting, main_quit does not exit program,
		# it keeps printing warnings,
		# and makes you close the terminal or force kill the proccess
		sys.exit(0)

	def consoleClearButtonClicked(self, widget=None):
		self.consoleTextview.get_buffer().set_text("")

	def verbosityComboChanged(self, widget=None):
		verbosity = self.verbosityCombo.get_active()
		# or int(self.verbosityCombo.get_active_text())
		log.setVerbosity(verbosity)

	def convertClicked(self, widget=None):
		inPath = self.convertInputEntry.get_text()
		if not inPath:
			self.status("Input file path is empty!")
			log.critical("Input file path is empty!")
			return
		inFormat = self.convertInputFormatCombo.getActive()
		if inFormat:
			inFormatDesc = Glossary.formatsDesc[inFormat]
		else:
			inFormatDesc = ""
			# log.critical("Input format is empty!");return

		outPath = self.convertOutputEntry.get_text()
		if not outPath:
			self.status("Output file path is empty!")
			log.critical("Output file path is empty!")
			return
		outFormat = self.convertOutputFormatCombo.getActive()
		if outFormat:
			outFormatDesc = Glossary.formatsDesc[outFormat]
		else:
			outFormatDesc = ""
			# log.critical("Output format is empty!");return

		while gtk.events_pending():
			gtk.main_iteration_do(False)

		self.convertButton.set_sensitive(False)
		self.progressTitle = "Converting"
		readOptions = self.convertInputFormatCombo.optionsValues
		writeOptions = self.convertOutputFormatCombo.optionsValues
		try:
			# if inFormat=="Omnidic":
			#	dicIndex = self.xml.get_widget("spinbutton_omnidic_i")\
			#		.get_value_as_int()
			#	ex = self.glos.readOmnidic(inPath, dicIndex=dicIndex)
			# else:
			log.debug("readOptions: %s" % readOptions)
			log.debug("writeOptions: %s" % writeOptions)
			finalOutputFile = self.glos.convert(
				inPath,
				inputFormat=inFormat,
				outputFilename=outPath,
				outputFormat=outFormat,
				readOptions=readOptions,
				writeOptions=writeOptions,
			)
			if finalOutputFile:
				self.status("Convert finished")
			else:
				self.status("Convert failed")
			return bool(finalOutputFile)

		finally:
			self.convertButton.set_sensitive(True)
			self.assert_quit = False
			self.progressTitle = ""

		return True

	def convertInputEntryChanged(self, widget=None):
		inPath = self.convertInputEntry.get_text()
		inFormat = self.convertInputFormatCombo.getActive()
		if inPath.startswith("file://"):
			inPath = urlToPath(inPath)
			self.convertInputEntry.set_text(inPath)

		inExt = getCopressedFileExt(inPath)
		inFormatNew = Glossary.extFormat.get(inExt)

		if not inFormatNew:
			return

		if not isfile(inPath):
			return

		if self.pref["ui_autoSetFormat"] and not inFormat:
			self.convertInputFormatCombo.setActive(inFormatNew)

		self.status("Select output file")
		if self.pref["ui_autoSetOutputFileName"]:
			outFormat = self.convertOutputFormatCombo.getActive()
			outPath = self.convertOutputEntry.get_text()
			if outFormat:
				if not outPath and "." in inPath:
					outPath = splitext(inPath)[0] + \
						Glossary.formatsExt[outFormat][0]
					self.convertOutputEntry.set_text(outPath)
					self.status("Press \"Convert\"")

	def convertOutputEntryChanged(self, widget=None):
		outPath = self.convertOutputEntry.get_text()
		outFormat = self.convertOutputFormatCombo.getActive()
		if not outPath:
			return
		# outFormat = self.combobox_o.get_active_text()
		if outPath.startswith("file://"):
			outPath = urlToPath(outPath)
			self.convertOutputEntry.set_text(outPath)

		if self.pref["ui_autoSetFormat"] and not outFormat:
			outExt = getCopressedFileExt(outPath)
			try:
				outFormatNew = Glossary.extFormat[outExt]
			except KeyError:
				pass
			else:
				self.convertOutputFormatCombo.setActive(outFormatNew)

		if self.convertOutputFormatCombo.getActive():
			self.status("Press \"Convert\"")
		else:
			self.status("Select output format")

	def reverseLoad(self):
		pass

	def reverseStartLoop(self):
		pass

	def reverseStart(self):
		if not self.reverseLoad():
			return
		###
		self.reverseStatus = "doing"
		self.reverseStartLoop()
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(True)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(True)

	def reverseStartClicked(self, widget=None):
		self.waitingDo(self.reverseStart)

	def reversePause(self):
		self.reverseStatus = "pause"
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(False)
		self.reverseResumeButton.set_sensitive(True)
		self.reverseStopButton.set_sensitive(True)

	def reversePauseClicked(self, widget=None):
		self.waitingDo(self.reversePause)

	def reverseResume(self):
		self.reverseStatus = "doing"
		###
		self.reverseStartButton.set_sensitive(False)
		self.reversePauseButton.set_sensitive(True)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(True)

	def reverseResumeClicked(self, widget=None):
		self.waitingDo(self.reverseResume)

	def reverseStop(self):
		self.reverseStatus = "stop"
		###
		self.reverseStartButton.set_sensitive(True)
		self.reversePauseButton.set_sensitive(False)
		self.reverseResumeButton.set_sensitive(False)
		self.reverseStopButton.set_sensitive(False)

	def reverseStopClicked(self, widget=None):
		self.waitingDo(self.reverseStop)

	def reverseInputEntryChanged(self, widget=None):
		inPath = self.reverseInputEntry.get_text()
		inFormat = self.reverseInputFormatCombo.getActive()
		if inPath.startswith("file://"):
			inPath = urlToPath(inPath)
			self.reverseInputEntry.set_text(inPath)

		inExt = getCopressedFileExt(inPath)
		inFormatNew = Glossary.extFormat.get(inExt)

		if inFormatNew and self.pref["ui_autoSetFormat"] and not inFormat:
			self.reverseInputFormatCombo.setActive(inFormatNew)

		if self.pref["ui_autoSetOutputFileName"]:
			outExt = ".txt"
			outPath = self.reverseOutputEntry.get_text()
			if inFormatNew and not outPath:
				outPath = splitext(inPath)[0] + "-reversed" + outExt
				self.reverseOutputEntry.set_text(outPath)

	def reverseOutputEntryChanged(self, widget=None):
		pass

	def progressInit(self, title):
		self.progressTitle = title

	def progress(self, rat, text=None):
		if not text:
			text = "%%%d" % (rat*100)
		text += " - %s" % self.progressTitle
		self.progressBar.set_fraction(rat)
		# self.progressBar.set_text(text)  # not working
		self.status(text)
		while gtk.events_pending():
			gtk.main_iteration_do(False)



# -*- coding: utf-8 -*-

import re

from typing import (
	Optional,
	List,
)

from .text_utils import (
	fixUtf8,
)

from .entry_base import BaseEntry
from .glossary import Glossary



class EntryFilter(object):
	name = ""
	desc = ""

	def __init__(self, glos: Glossary):
		self.glos = glos

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		"""
			returns an Entry object, or None to skip
				may return the same `entry`,
				or modify and return it,
				or return a new Entry object
		"""
		return entry


class StripEntryFilter(EntryFilter):
	name = "strip"
	desc = "Strip Whitespaces"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		entry.strip()
		entry.replace("\r", "")
		return entry


class NonEmptyWordFilter(EntryFilter):
	name = "non_empty_word"
	desc = "Non-empty Words"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		if not entry.getWord():
			return
#		words = entry.getWords()
#		if not words:
#			return
#		wordsStr = "".join([w.strip() for w in words])
#		if not wordsStr:
#			return
		return entry


class NonEmptyDefiFilter(EntryFilter):
	name = "non_empty_defi"
	desc = "Non-empty Definition"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		if not entry.getDefi():
			return
		return entry


class RemoveEmptyAndDuplicateAltWords(EntryFilter):
	name = "remove_empty_dup_alt_words"
	desc = "Remove Empty and Duplicate Alternate Words"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		entry.removeEmptyAndDuplicateAltWords()
		if not entry.getWords():
			return
		return entry


class FixUnicodeFilter(EntryFilter):
	name = "fix_unicode"
	desc = "Fix Unicode"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		entry.editFuncWord(fixUtf8)
		entry.editFuncDefi(fixUtf8)
		return entry


class LowerWordFilter(EntryFilter):
	name = "lower_word"
	desc = "Lowercase Words"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		entry.editFuncWord(str.lower)
		return entry


class RemoveHtmlTagsAll(EntryFilter):
	name = "remove_html_all"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		from bs4 import BeautifulSoup

		def fixStr(st: str) -> str:
			return BeautifulSoup(st, "lxml").text

		entry.editFuncDefi(fixStr)
		return entry


class RemoveHtmlTags(EntryFilter):
	name = "remove_html"

	def __init__(self, glos: Glossary, tags: List[str]):
		self.glos = glos
		self.tags = tags

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		import re

		def fixStr(st: str) -> str:
			tagsRE = "|".join(self.tags)
			pattern = f"<.?({tagsRE})[^>]*>"
			st = re.sub(pattern, "", st)
			return st

		entry.editFuncDefi(fixStr)
		return entry


class SkipDataEntryFilter(EntryFilter):
	name = "skip_resources"
	desc = "Skip Resources"

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		if entry.isData():
			return
		return entry


class LangEntryFilter(EntryFilter):
	name = "lang"
	desc = "Language-dependent Filters"

	def run_fa(self, entry: BaseEntry) -> Optional[BaseEntry]:
		from pyglossary.persian_utils import faEditStr
		entry.editFuncWord(faEditStr)
		entry.editFuncDefi(faEditStr)
		# RLM = "\xe2\x80\x8f"
		# defi = "\n".join([RLM+line for line in defi.split("\n")])
		# for GoldenDict ^^ FIXME
		return entry

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		langs = (
			self.glos.getInfo("sourceLang") +
			self.glos.getInfo("targetLang")
		).lower()
		if "persian" in langs or "farsi" in langs:
			entry = self.run_fa(entry)

		return entry


class CleanEntryFilter(EntryFilter):  # FIXME
	name = "clean"
	desc = "Clean"

	def cleanDefi(self, st: str) -> str:
		st = st.replace("♦  ", "♦ ")
		st = re.sub("[\r\n]+", "\n", st)
		st = re.sub(" *\n *", "\n", st)

		"""
		This code may correct snippets like:
		- First sentence .Second sentence. -> First sentence. Second sentence.
		- First clause ,second clause. -> First clause, second clause.
		But there are cases when this code have undesirable effects
		( "<" represented as "&lt;" in HTML markup):
		- <Adj.> -> < Adj. >
		- <fig.> -> < fig. >
		"""
		"""
		for j in range(3):
			for ch in ",.;":
				st = replacePostSpaceChar(st, ch)
		"""

		st = re.sub("♦\n+♦", "♦", st)
		if st.endswith("<p"):
			st = st[:-2]
		st = st.strip()
		if st.endswith(","):
			st = st[:-1]

		return st

	def run(self, entry: BaseEntry) -> Optional[BaseEntry]:
		entry.editFuncDefi(self.cleanDefi)
		return entry



from formats_common import *

import sys
import os
from os.path import (
	join,
	split,
	splitext,
	isfile,
	isdir,
	exists,
)

from typing import (
	Optional,
	Any,
	AnyStr,
	Tuple,
	List,
	Dict,
	Iterator,
)

import logging

log = logging.getLogger("root")

from pprint import pformat

from paths import rootDir
sys.path.insert(0, rootDir)

from pyglossary.flags import *

from pyglossary import core
from pyglossary.option import *
from pyglossary.file_utils import FileLineWrapper
from pyglossary.text_utils import toStr, toBytes
from pyglossary.os_utils import indir
from pyglossary.entry_base import BaseEntry

from pyglossary.glossary_type import GlossaryType

enable = False
format = "Unknown"
description = "Unknown"
extensions = [] # type: List[str]

# key is option/argument name, value is instance of Option
optionsProp = {} # type: Dict[str, Option]

depends = {}
supportsAlternates = False
sortOnWrite = DEFAULT_NO
sortKey = None



# -*- coding: utf-8 -*-

from .text_utils import replacePostSpaceChar

def faEditStr(st: str) -> str:
	return replacePostSpaceChar(
		st.replace("ي", "ی")
		  .replace("ك", "ک")
		  .replace("ۂ", "هٔ")
		  .replace("ہ", "ه"),
		"،",
	)
